//*****************************************************************************
//*****************************************************************************
// OFORTH : NATIVE FUNCTIONS
//*****************************************************************************
//*****************************************************************************

/*
  Copyright (c) 2014, 2018 Franck Bensusan. All rights reserved.

  This file is part of Oforth sources engine.

  Oforth is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation, either version 3
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, see http://www.gnu.org/licenses/.
*/

// ****************************************************************************


// ****************************************************************************
// Note : The integer division algorithm part has been adapted from 
// Per M.A. Bothner work on Kawa.
// ****************************************************************************


#include "oforth.h"

#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/stat.h>
#include <unistd.h>
#include <math.h>

extern Attribute *add_attribute(CThread *, Class *, MySymbol *, MyProperty *, vint);
extern Class *class(CThread *, char *, Class *);
extern MyConstant *constant(CThread *, char *, Cell);
extern MyVariable *variable(CThread *thread, char *name);
extern MyMethod *add_meth(CThread *thread, MyMessage *m,Class *, Class *, fptr, Class *);

extern MySymbol *find_symbol(char *name, vint hash);
extern MySymbol *add_symbol(CThread *, char *);
extern MyPackage *currentPackage(CThread *thread);
extern MyPackage *rootPackage(CThread *thread);
extern MyWord *wordFind(CThread *thread, uchar *name);
extern MyWord *wordPackageFind(CThread *thread, MyPackage *pack, MySymbol *name, vint hash);
extern void addWordToCurrent(CThread *thread, MyWord *w);

extern PCell allocRawMem(CThread *thread, vint n);
extern PCell allocDictCell(CThread *thread, vint n);
extern MyObject *allocDictObject(CThread *thread, Class *cl, vint n);
extern MyString *allocDictString(CThread *, char *, vint);
extern MyArray *allocDictArray(CThread *thread, vint size);
extern void checkImmutable(register CThread *thread, MyObject *o);
extern void setObjPersistent(CThread *, MyObject *o);
extern Attribute *create_att(CThread *thread, Implementor *impl, MySymbol *name, MyProperty *p, vint mutable);

extern void *allocMem(CThread *, vint, char **);
extern void freeMem(CThread *, PTR);
extern void decimalValues(CThread *thread, MyNatural *n, vint *res, int *size);
extern int totalHeap();

extern Class *objectClass(MyObject *o);
extern vint isKindOf(MyObject *o, Class *parent);

extern vint hash_value(char *);
extern void call_poly();
extern void closureSetAtt();
extern void pushMarkStack(CThread *, MyObject *);

extern void sysExceptionFromC(CThread *thread, char *s);
extern void sysExParamFromC(CThread *thread, char *s, void *x);
extern void outOfMemory(CThread *thread);
extern MyMessage *selector(CThread *thread, char *mn);

extern Cell getvar(CThread *thread, MyVariable *v);
extern void setvar(CThread *thread, MyVariable *v, Cell value);

extern Class *Object, *Word, *Classe, *Function, *Message, *Block;
extern Class *Method, *Abstract, *OSLib, *OSProc, *Property, *Package;
extern Class *Buffer, *Symbol, *String, *Array;
extern Class *File, *Integer, *Float, *Exception;

extern vint genericVT;
extern Cell *allNewVT;
extern Class **allClasses;
extern vint nbClasses;
extern MyPackage *oforthPackage;
extern MyWord *lastWord;

extern CThread *mainThread;

extern MyArray *allPath;
extern vint nPath;
extern uchar InputChar[0x80];

extern MyString *nextWord;
extern MyVariable *allBlocks;
extern MyVariable *CS;
extern MyVariable *STATE, *CURRENT;
extern MyVariable *ICODE;
extern MyVariable *LASTCALL, *CPBEFORELASTCALL, *CPLASTCALL;
extern MyVariable *SOURCE, *SOURCETYPE, *SOURCENAME, *SOURCELINE, *SOURCEINDEX;

extern vint codeMem;
extern int tickInit;

#ifdef __DBG__
extern int pwo;
#endif

uchar *targetCode(CThread *thread);
vint utf8ToChars(register CThread *thread, register vint c);


//*****************************************************************************
// Compilation
//*****************************************************************************

//*****************************************************************************
void compileError() 
//*****************
{
   printf("Serious compilation error, leaving..."); exit(KO);
}
//*****************************************************************************
vint indexCode(CThread *thread)
//*****************************
{
vint icode;

   icode = (vint) TOCINT(getvar(thread, ICODE));
   if (icode > WARNCODE) sysExceptionFromC(thread, "Definition is too long");
   return icode;
}
//*****************************************************************************
uchar *targetCode(CThread *thread)
// VALIDATED *********************
{
MyTask *t;
CodePage *cp, *cpprev;

   t = thread->currentTask;
   cp = cpprev = t->codePage;

   if (cp == NULL || (OPAGESIZE - (cp->target - (uchar *) cp)) <= MAXCODE) {
      cp = t->codePage = AllocPage;
      if (cp == NULL) outOfMemory(thread);
      LockAdd(codeMem, OPAGESIZE); 
      cp->prev = cpprev;
      cp->target = cp->code;
      if (PageProtectEx(cp) == PageProtError) compileError();
      }
   return cp->target;
}
//*****************************************************************************
void initCode(CThread *thread)
//****************************
{
MyTask *t;

   t = thread->currentTask;
   if (t->codeBuffer == NULL) 
       t->codeBuffer = (uchar *) allocRawMem(thread, MAXCODE); 

   setvar(thread, ICODE , (Cell) OZERO);
}
//*****************************************************************************
void closureOffsetUsed(MyBlock *bl, vint offset)
//**********************************************
{
   if (bl->nbVars < offset) bl->nbVars = offset;
   bl->varused |= (1 << offset );
}
//*****************************************************************************
void N_setLongJump(CThread *thread)
//*********************************
{
TryNode *tr;

   POPDS(tr, TryNode *);
   tr->whenAddr = thread->currentTask->codePage->target + indexCode(thread);
}
//*****************************************************************************
void N_closureVarUsed(CThread *thread)
//************************************
{
vint slots, offset, state, i;
MyBlock *bl;
MyArray *blocks;

   POPDS(slots, vint);  slots  = TOCINT(slots);
   POPDS(offset, vint); offset = TOCINT(offset);

   blocks = (MyArray *) getvar(thread, allBlocks);
   
   for(state = TOCINT(getvar(thread, STATE)); state > 1; state--) {
      bl = (MyBlock *) (blocks->buf)[state - 1];

      for(i = 0; i < slots; i++) closureOffsetUsed(bl, offset + i);
      }
}
//*****************************************************************************
void N_BYTECODE(CThread *thread)
//******************************
{
vint icode;
uchar *code;
vint byte;

   icode = indexCode(thread);
   code = thread->currentTask->codeBuffer + icode;
   byte = TOCINT(TOP); DROP;
   *code = byte;
  setvar(thread, ICODE, (Cell) TOOINT(icode + 1)); 
}
//*****************************************************************************
void N_RAWCODE(CThread *thread)
//*****************************
{
vint icode;
uchar *code;
Cell value;

   icode = indexCode(thread);
   code = thread->currentTask->codeBuffer + icode;
   value = (Cell) TOCINT(TOP); DROP;
   *((Cell *) (code)) = value; 
   setvar(thread, ICODE, (Cell) TOOINT(icode + sizeof(Cell))); 
}
//*****************************************************************************
void N_CELLCODE(CThread *thread)
//******************************
{
vint icode;
uchar *code;
Cell value;

   icode = indexCode(thread);
   code = thread->currentTask->codeBuffer + icode;
   POPDS(value, Cell)
   *((Cell *) (code)) = value; 
   setvar(thread, ICODE, (Cell) TOOINT(icode + sizeof(Cell))); 
}
//*****************************************************************************


#if defined(__x8632__)

//*****************************************************************************
int dangerousLiteral(Cell value)
//******************************
{
int i;
uchar *pval = (uchar *) &value, c;

   for(i = 0; i < sizeof(Cell); i++) {
      c = pval[i];
      if (c == 0xC2 || c == 0xC3 || c == 0xCA || c == 0xCB) return OK;
      }
   return KO;
}
//*****************************************************************************
void asm_param(CThread *thread)
//*****************************
{
    PUSHDS(TOOINT(0xBD)); N_BYTECODE(thread); N_CELLCODE(thread);
}
//*****************************************************************************
void asm_tos(CThread *thread)
//***************************
{
    PUSHDS(TOOINT(0xBB)); N_BYTECODE(thread); N_CELLCODE(thread);
}
//*****************************************************************************
void asm_call(CThread *thread)
//****************************
{
vint icode;
uchar *code, *addr;
MyTask *t;

   icode = indexCode(thread);
   setvar(thread, CPBEFORELASTCALL, (Cell) TOOINT(icode)); 
   
   t = thread->currentTask;
   code = t->codeBuffer + icode;
   POPDS(addr, uchar *);
   setvar(thread, LASTCALL, (Cell) (addr+1)); 
   
   *code = 0xE8;
   *((vint *) (code + 1)) = addr - icode - t->codePage->target - sizeof(Cell) - 1;
   icode += sizeof(Cell) + 1; 
   setvar(thread, ICODE, (Cell) TOOINT(icode)); 
   setvar(thread, CPLASTCALL, (Cell) TOOINT(icode)); 
   // dbg("icode = %d\n", TOOINT(icode));
}
//*****************************************************************************
void asm_compile(CThread *thread)
//*******************************
{
   asm_param(thread); asm_call(thread);
}
//*****************************************************************************
void N_RET(CThread *thread)
//*************************
{
vint icode;
uchar *code, *addr;
MyTask *t;

   icode = indexCode(thread);

   // If last instruction is a call, tail optimization
   if ((vint) getvar(thread, CPLASTCALL) == TOOINT(icode)) {
      icode = TOCINT(getvar(thread, CPBEFORELASTCALL));
      t = thread->currentTask;
      code = t->codeBuffer + icode;
      addr = ((uchar *) getvar(thread, LASTCALL)) - 1;
      icode += sizeof(Cell) + 1; 
      *code = 0xE9;
      *((vint *) (code + 1)) = addr - icode - t->codePage->target;
      setvar(thread, ICODE, (Cell) TOOINT(icode)); 
      }
   else { PUSHDS(TOOINT(0xC3)); N_BYTECODE(thread); }
}
//*****************************************************************************
void N_lit(CThread *thread)
//*************************
{
uchar *code, *buf;
Cell value;
int offset;

    buf = thread->currentTask->codeBuffer;
    code = buf + indexCode(thread);
    offset = (int) TOP;

    value = (Cell) SECOND;
    setObjPersistent(thread, (MyObject *) value); 

    if (offset == 0) {
       offset = OZERO;
       *code++ = 0x83;               // sub $4, EDI
       *code++ = 0xEF;
       *code++ = 0x04;
       }

    offset = TOCINT(offset) * sizeof(Cell);
    if (offset != 0) offset = 0x100 - offset;

    if (dangerousLiteral(value) == KO) {
       *code++ = 0xC7;               // mov value, (EDI)
       if (offset == 0) { *code++ = 0x07; } 
       else { *code++ = 0x47; *code ++= offset; }
       *((Cell *) (code)) = value; code += sizeof(Cell); 
       }
    else {                           // warning : possible ret opcode into literal
       *code++ = 0xBD;
       value = (Cell) (((int) value) ^ tickInit); *((Cell *) (code)) = value;
       code += sizeof(Cell);
       *code++ = 0x81;              
       *code++ = 0xF5;
       value = (Cell) tickInit;
       *((Cell *) (code)) = value; code += sizeof(Cell); 

       *code++ = 0x89;              
       if (offset == 0) { *code++ = 0x2F; } 
       else { *code++ = 0x6F; *code++ = offset; }
       }

    setvar(thread, ICODE, (Cell) TOOINT(code - buf)); 
    DROP2;
}
//*****************************************************************************
char *asmContinue = "\xE9\x01\x00\x00\x00\xBD\xFA\xFF\xFF\x3F";
char* asmBreak    = "\xE9\x01\x00\x00\x00\xBD\xFE\xFF\xFF\x3F";

//*****************************************************************************
void N_resolveContinue(CThread *thread)
//*************************************
{
uchar *jumpAddr, *addr, *code, *buf;

   buf = thread->currentTask->codeBuffer;
   code     = buf + indexCode(thread);
   addr     = buf + TOCINT(TOP); DROP;
   jumpAddr = buf + TOCINT(TOP); DROP;
   
   for(; addr < code; addr++) {
      if (memcmp((char *) addr, asmContinue, 10) == OK) {
          *((int *)  (addr + 1)) = jumpAddr - addr - sizeof(Cell) - 1;
          *((Cell *) (addr + 6)) = 0;
         addr += 10;
         }
      }
}
//*****************************************************************************
void N_resolveBreak(CThread *thread)
//**********************************
{
uchar *addr, *code, *buf;

   buf = thread->currentTask->codeBuffer;
   code = buf + indexCode(thread);
   addr = buf + TOCINT(TOP); DROP;
   
   for(; addr < code; addr++) {
      if (memcmp((char *) addr, asmBreak, 10) == OK) {
          *((int *)  (addr + 1)) = code - addr - sizeof(Cell) - 1;
          *((Cell *) (addr + 6)) = 0;
         addr += 10;
         }
      }
}
//*****************************************************************************

#endif

//*****************************************************************************
void N_closureVarSet(CThread *thread)
//***********************************
{
vint offset, state;
MyBlock *bl;
MyArray *blocks;

   POPDS(offset, vint);
   offset = sizeof(MyClosure) + (TOCINT(offset) - 2) * sizeof(Cell); 
   blocks = (MyArray *) getvar(thread, allBlocks);

   PUSHDS(offset); asm_param(thread); PUSHDS(closureSetAtt); asm_call(thread);
   setvar(thread, CPLASTCALL, NULL); 

   for(state = TOCINT(getvar(thread, STATE)); state > 1; state--) {
      bl = (MyBlock *) (blocks->buf)[state - 1];
      bl->objptr |= MSKMUTABLE;
      }
}
//*****************************************************************************


//*****************************************************************************
// Polymorphism
//*****************************************************************************

//*****************************************************************************
MyMethod *implementFor(register MyMethod *node, register Class *cl)
// ****************************************************************
{
   while( node != NULL && node->implementor != cl ) node = node->next;
   return node;
}
//*****************************************************************************
MyMethod *findClassMethod(MyMessage *m, register Class *cl)
// *********************************************************
{
MyMethod *n = NULL, *p;

   p = m->classeList;
   while( cl != NULL && (n = implementFor(p, cl)) == NULL) cl = cl->super;
   return n;
}
//*****************************************************************************
MyMethod *findMethod(MyMessage *m, register Class *cl)
// VALIDATED ********************************************
{
MyMethod *n, *objMethod, *hierarchyImpl = NULL;
register MyNodeProp *np;
Class *cl1;

   // dbg("m = %x, name = %s cl = %s \n", m, symbuf(m->name), symbuf(cl->name));
   objMethod = m->objectList;
   cl1 = cl;
   
   // First search into the class hierarchy
   while(cl1 != NULL && (hierarchyImpl = implementFor(objMethod, cl1)) == NULL) 
       cl1 = cl1->super;

   if (objectClass((MyObject *) hierarchyImpl) == Method) return hierarchyImpl;

   // If implement not found, search hierarchy with properties
   while(cl != NULL) {
      if ((n = implementFor(objMethod, cl)) != NULL) return n;
      np = cl->propList;
      while(np != NULL) {
         if ((n = implementFor(objMethod, (Class *) np->prop)) != NULL) return n;
         np = np->next;
         }
      cl = cl->super;
      }

   return NULL;
}
//*****************************************************************************
void polymorph(register CThread *thread)
// VALIDATED ***************************
{
MyMessage *m;
PCell vt;
vint classId;
register Class *cl;
register MyMethod *n;

   POPDS(vt, PCell); POPDS(classId, vint); m = (MyMessage *) TOP;
   cl = (Class *) allClasses[classId];

   if (vt == m->vt) { n = findMethod(m, cl); }
   else {
      n = findClassMethod(m, cl);
      if (n == NULL) n = findMethod(m, Classe);
      }

   if (n != NULL) { vt[classId] = (Cell) n->code; }
   PUSHDS(n);
}
//*****************************************************************************
void _reallocVT(register CThread *thread)
//***************************************
{
MyMessage *m;
vint size;
Cell *oldvt, *newvt;
Cell *max;
register vint i, newsize, oldsize;

   POPDS(size, vint); m = (MyMessage *) TOP;

   LockRes(thread->vm);
       oldvt = m->vt;
       oldsize = (vint) *oldvt;
       if (oldsize >= size) goto LReallocReturn;

       newvt = allocMem(thread, (size + 1) * sizeof(Cell), (char **) &max); 
       if (oldsize != 0) freeMem(thread, oldvt);

       newvt[0] = (Cell) (max - newvt - 1); 
       newsize = (vint) newvt[0];
       for(i = 1; i <= newsize; i++) newvt[i] = (Cell) call_poly;
       m->vt = newvt; 
LReallocReturn:
   RelRes(thread->vm);
   
   // dbg("New vt (method %x) : old = %d, new = %d\n", symbuf(m->name), oldsize, newsize);
}   
//*****************************************************************************


//*****************************************************************************
// INPUT & PARSING
//*****************************************************************************

//*****************************************************************************
void ungetChar(register CThread *thread, Cell source, Cell type, register char c)
// VALIDATED ********************************************************************
{
FILE *f;
vint index;

    index = TOCINT((vint) getvar(thread, SOURCEINDEX)) - 1;
    setvar(thread, SOURCEINDEX, (Cell) TOOINT(index));

    if (type == (Cell) String) *(((MyString *) source)->buf + index) = c;
    else { f = (FILE *) OINTTOF(source); ungetc(c, f); }
   
    if (c == '\n' || c == '\r') 
        setvar(thread, SOURCELINE, (Cell) (((vint) getvar(thread, SOURCELINE)) - 2));
}
//*****************************************************************************
int getByte(CThread *thread, Cell source, Cell type)
// VALIDATED ****************************************
{
register int n;
vint index;

    // dbg("source = %x, type = %x String = %x\n", source, type, String);

    if (type == (Cell) String) { 
        index = TOCINT(getvar(thread, SOURCEINDEX));
        n = (int) *(((MyString *) source)->buf + index); 
        setvar(thread, SOURCEINDEX, (Cell) TOOINT((index + 1)));
        }
    else { n = fgetc((FILE *) OINTTOF(source)); }

   if (n == '\r') {
      n = getByte(thread, source, type);
      if (n != '\n') { ungetChar(thread, source, type, n); n = '\n'; }
      }
      
   if (n == '\n' && type != (Cell) String) { 
        setvar(thread, SOURCELINE, (Cell) TOOINT(TOCINT(getvar(thread, SOURCELINE)) + 1));
        setvar(thread, SOURCEINDEX, (Cell) OZERO);
       }
   return n;
}
//*****************************************************************************
vint getChar(CThread *thread, Cell source, Cell type)
//***************************************************
{
register int c;
register uchar d;
register vint size, translate, i;

   c = getByte(thread, source, type);
   // dbg("c=<%d> ", c);

   if (c <= 0x7F || c == (unsigned char) EOF) return c;
   if (c >= 0xC2 && c <= 0xDF)        { size = 1; translate = 0x0003080; }
   else { if (c >= 0xE0 && c <= 0xEF) { size = 2; translate = 0x00E2080; }
   else { if (c >= 0xF0 && c <= 0xF4) { size = 3; translate = 0x3C82008; }
   else { return '?'; } } }

   for(i = 1; i <= size; i++) {
      d = getByte(thread, source, type);
      if (d < 0x80 || d > 0xBF) { return '?'; }
      if (i == 1) {
         if (size == 2) {
            if (c == 0xE0 && d < 0xA0) return '?';
            if (c == 0xED && d > 0x9F) return '?';
            }
         if (size == 3) {
            if (c == 0xF0 && d < 0x90) return '?';
            if (c == 0xF4 && d > 0x8F) return '?';
            }
         }
      c = (c << 6) + d;
      }
   c -= translate;
   return c;
}
//*****************************************************************************
MYINT getEscapeSeq(CThread *thread, Cell source, Cell type)
//**********************************************************
{
register vint c;
vint d = 0, i;

   c = getByte(thread, source, type);
   switch(c) {
      case 'n' : c = '\n'; break;
      case 't' : c = '\t'; break;
      case 'r' : c = '\r'; break;
      case 'b' : c = '\b'; break;

      case 'u' :
      case 'U' :
         for(i = 0; i < 4; i++) {
            c = getByte(thread, source, type);
            // dbg("c = %c d= %d\n", c, d);
            if (c >= '0' && c <= '9') { d = d * 16 + (c - '0'); continue; }
            if (c >= 'a' && c <= 'f') { d = d * 16 + (c - 'a' + 10); continue; }
            if (c >= 'A' && c <= 'F') { d = d * 16 + (c - 'A' + 10); continue; } 
            goto LEscapeError;
            }
         if (d >= 0xD800 && d <= 0xDFFF) goto LEscapeError;
         c = d;
         break;

      default:
         break;
      }

   return c;

LEscapeError:
   sysExceptionFromC(thread, "Wrong character");
   return 0;
}
//*****************************************************************************
void N_parse_token(CThread *thread)
// VALIDATED ********************
{
register int c, first = 0;
uchar *wt, *word;
Cell source, type;
   
   source = getvar(thread, SOURCE);
   type   = getvar(thread, SOURCETYPE);

   wt = word = nextWord->buf; 

   do { c = getChar(thread, source, type); }
   while(c == ' ' || c == '\t' || c == '\n' || c == '\r');

   if (c != EOF && c != '\0') { 

      if (c < 0x80 ) {
         if (c == '0' || c == '-' || c == '#') first = c;
         else if (InputChar[c] & CPREFIX) goto LLastChar;
         }

      while( 1 ) {
         *wt++ = c;
         if (wt - nextWord->buf >= MAXNAME) {
             *wt = '\0'; 
              sysExceptionFromC(thread, "Name is too long");
              }

         c = getChar(thread, source, type);
         if (first != 0) {
             if (first == '0' && (c == 'x' || c == 'b')) goto LLastChar;
             if (first == '-' && c == '>') goto LLastChar;
             if (first == '#') {
                 if (c == '!' || c == '[') goto LLastChar;
                 break;
                 }
             first = 0;
             }
         if( c == EOF) break;
         if (c < 0x80 && (InputChar[c] & CSUFFIX)) break;
         }
      }

    ungetChar(thread, source, type, c); goto LWordFound;

LLastChar:
    *wt++ = c;

LWordFound:
    *wt = '\0'; nextWord->free1 = wt;

#ifdef __DBG__
   if (pwo == OK) printf("%s ", nextWord->buf);
#endif
}
//*****************************************************************************
void N_parsechar(CThread *thread)
//*******************************
{
vint c;
Cell source, type;

   source = getvar(thread, SOURCE);
   type   = getvar(thread, SOURCETYPE);

   c = getChar(thread, source, type);

   if (c == EOF || c == 0) {
        ungetChar(thread, source, type, c);
        c = nil;
       }
   else { c = (vint) TOOINT(c); }

   PUSHDS(c);
}
//*****************************************************************************
void N_nextchar(CThread *thread)
//******************************
{
register char c;
Cell source, type;

   source = getvar(thread, SOURCE);
   if (source == NULL) { PUSHDS(NULL); return; }

   type   = getvar(thread, SOURCETYPE);

   do { c = getChar(thread, source, type); } 
   while(c == ' ' || c == '\t' || c == '\n' || c == '\r');

   ungetChar(thread, source, type, c);
   PUSHDS(TOOINT(c));
}
//*****************************************************************************
void N_char(CThread *thread)
//**************************
{
vint c, d;
Cell source, type;
   
   source = getvar(thread, SOURCE);
   type   = getvar(thread, SOURCETYPE);

   c = getChar(thread, source, type);
   if (c == '\\') c = getEscapeSeq(thread, source, type); 
   
   d = getChar(thread, source, type);
   if (d != '\'') sysExceptionFromC(thread, "Unterminated character");  

   // dbg("Character read = %d %c\n", c, c);
   PUSHDS(TOOINT(c));
}
//*****************************************************************************
char *readString(CThread *thread)
//*******************************
{
vint c, n;
char *s;
Cell source, type;
   
   source = getvar(thread, SOURCE);
   type   = getvar(thread, SOURCETYPE);

   s = (char *) nextWord->buf; 

   while(1) {
      c = getChar(thread, source, type); 
      if (c == '\"') break;
      if (c == EOF || c == 255 || c == 0) 
         sysExceptionFromC(thread, "Unterminated string"); 

      if (c == '\\') { c = getEscapeSeq(thread, source, type); }
      if (c < 0x80) { *s++ = c; continue; }
      n = utf8ToChars(thread, c); 
      strcpy(s, (char *) FORMAT);
      s += n;
      }

    *s = 0;
    return s;
}
//*****************************************************************************
void N_dblQuote(CThread *thread)
//******************************
{
    PUSHDS(TOOINT(readString(thread) - (char *) nextWord->buf));
}
//*****************************************************************************
void N_dblQuoteCompile(CThread *thread)
//*************************************
{
    readString(thread);
    PUSHDS(allocDictString(thread, (char *) nextWord->buf, OK));
}
//*****************************************************************************
void N_comment(CThread *thread)
// ****************************
{
char c;
Cell source, type;
   
   source = getvar(thread, SOURCE);
   type   = getvar(thread, SOURCETYPE);

   do { c = getChar(thread, source, type); } 
   while (c != '\n' && c != EOF && c != '\0' ); 

   ungetChar(thread, source, type, c);
}
//*****************************************************************************
void N_mem(register CThread *thread)
//**********************************
{
CodePage *cp;
vint freeCode;
int heap;

   heap = totalHeap();
   cp = thread->currentTask->codePage;
   if (cp == NULL) { freeCode = 0; }
   else { freeCode = OPAGESIZE - (vint) cp->target + (vint) cp->code; }

   printf("Memory used\n- Code      : %8d Ko (%8d bytes free) \n",   
      codeMem / 1024, freeCode);

   printf("- Dict+Heap : %8d Ko (%8d bytes free)", 
       heap, thread->persistentFree + thread->currentTask->dpCellLeft * sizeof(Cell) );

   printf("\n- Total     : %8d Ko\n", heap + codeMem / 1024);
}
//*****************************************************************************


//*****************************************************************************
// WORDS and dictionary 
//*****************************************************************************

//****************************************************************************r
void sysExOS(register CThread *thread)
//************************************
{
   sysExParamFromC(thread, strerror(errno), (void *) TOOINT(errno));
}
//*****************************************************************************
void badParams(register CThread *thread)
//*************************************
{
   sysExceptionFromC(thread, "Bad parameters");
}
//*****************************************************************************
void badStack(register CThread *thread)
//*************************************
{
   sysExceptionFromC(thread, "Stack underflow");
}
//*****************************************************************************
void checkType(CThread *thread, void *o, Class *cl)
//*************************************************
{
   if (isKindOf((MyObject *) o, cl) == KO) 
      sysExParamFromC(thread, "Wrong parameter, required type is", cl->name);
}
//*****************************************************************************
void checkTypeInteger(CThread *thread, vint i)
//********************************************
{
   if (!ISINT(i)) 
      sysExceptionFromC(thread, "Wrong parameter, required type is small Integer");
}
//*****************************************************************************
void wordInit(register CThread *thread)
//*************************************
{
MyWord *w;
MyString *name;
MySymbol *s;
vint hash;

   w = (MyWord *) TOP;
   name = (MyString *) SECOND;
   checkType(thread, name, String);
   
   s = add_symbol(thread, (char *) name->buf);
   hash = hash_value((char *) name->buf) % HASH_SIZE;
   if (wordPackageFind(thread, currentPackage(thread), s, hash) != NULL) 
      sysExParamFromC(thread, "Can't redefine word", s);

   w->name = s;
   addWordToCurrent(thread, w);
   lastWord = w;
   DROP2
}
//*****************************************************************************
void N_Word_find(register CThread *thread)
// VALIDATED *****************************
{
MyString *s;

   s = (MyString *) TOP;
   checkType(thread, s, String);
   TOP = (Cell) wordFind(thread, s->buf); 
}
//*****************************************************************************
void N_Word_forget(register CThread *thread)
//******************************************
{
MyPackage *pack;
MyWord *word, *h, *hprev;
vint hash;
Cell *buf;

   if ((pack = currentPackage(thread)) == NULL) goto LForgetError;
   word = (MyWord *) TOP;
   hash = hash_value((char *) symbuf(word->name)) % HASH_SIZE;
   buf = pack->wrdlist->buf;
   h = hprev = (MyWord *) buf[hash];

   while(h != NULL && h != word) { hprev = h; h = h->next; }
   if (h == NULL) goto LForgetError;

   // h->name = NULL;
   if (h == hprev) { buf[hash] = (Cell) h->next; }
   else { hprev->next = h->next; }

   DROP
   return;

LForgetError:
   sysExceptionFromC(thread, "Can't forget word");

}
//*****************************************************************************
void N_Class_initialize(register CThread *thread)
// **********************************************
{
Class *cl, *super, **old;
vint i;

   CHECKSTACK(2)

   cl    = (Class *) TOP; 
   super = (Class *) SECOND; 
   checkType(thread, super, Classe);

   if (nbClasses >= MAX_CLASSES) 
      { sysExceptionFromC(thread, "Max number of classes reached"); }

   if (nbClasses == MIN_CLASSES - 1) {
      old = allClasses;
      allClasses = (Class **) allocDictCell(thread, MAX_CLASSES);
      memcpy(allClasses, old, MIN_CLASSES * sizeof(Cell));

      allNewVT    = (Cell *) allocDictCell(thread, MAX_CLASSES);
      allNewVT[0] = (Cell) (MAX_CLASSES - 1);
      for(i = 1; i < MAX_CLASSES; i++) allNewVT[i] = (Cell) call_poly;
      }
   
   nbClasses++; 
   allClasses[nbClasses] = cl;
   cl->super         = super; 
   cl->size          = super->size;
   cl->initFunction  = super->initFunction;
   cl->finalFunction = super->finalFunction;
   cl->classPtr      = nbClasses + (super->classPtr & MSKSTATE);
   cl->classPtr      &= ~MSKNOTATTGC;

   DROP2;
}
//*****************************************************************************
void N_Class_addAtt(register CThread *thread)
//*******************************************
{
MySymbol *name;
MyProperty *p;

   p = (MyProperty *) THIRD;
   name = (MySymbol *) FOURTH;
   checkType(thread, name, Symbol);
   add_attribute(thread, (Class *) TOP, name, p, (vint) SECOND);
   DROP4;
}
//*****************************************************************************
void N_Property_addAtt(register CThread *thread)
//*********************************************
{
MyProperty *p;
MySymbol *name;

   p = (MyProperty *) THIRD;
   name = (MySymbol *) FOURTH;
   checkType(thread, name, Symbol);
   create_att(thread, (Implementor *) TOP, name, p, (vint) SECOND);
   DROP4;
}
//*****************************************************************************
void N_Class_findAtt(register CThread *thread)
//********************************************
{
Class *cl;
MySymbol *name;
Attribute *na;
MyProperty *p;
vint index;

   cl = (Class *) TOP;
   p = (MyProperty *) SECOND;
   name = (MySymbol *) THIRD;
   index = cl->size - 1;

   while (cl != NULL) {
      na = cl->attList;
      while(na != NULL) {
         if (na->name == name && na->property == p) {
            THIRD = (Cell) na->isMutable;
            SECOND = (Cell) TOOINT(index); TOP = (Cell) OTRUE; 
            return;
            }
         index--;
         na = na->next;
         }
      cl = cl->super;
      }
   THIRD = (Cell) OFALSE;
   DROP2
}
//*****************************************************************************
void N_Property_findAtt(register CThread *thread)
//***********************************************
{
MyProperty *p;
MySymbol *name;
Attribute *na;

   p = (MyProperty *) TOP;
   name = (MySymbol *) THIRD;
   na = p->attList;

   while(na != NULL) {
      if(na->name == name) break;
      na = na->next;
      }

   if (na == NULL)
          { THIRD = (Cell) OFALSE; DROP2; }
   else   { THIRD = (Cell) na->isMutable; SECOND = (Cell) name; TOP = (Cell) OTRUE; } 
}
//*****************************************************************************
void N_Class_method(register CThread *thread)
// VALIDATED ********************************
{
MyMessage *m;

   m = (MyMessage *) SECOND; 
   checkType(thread, m, Message);

   SECOND = (Cell) findMethod(m, (Class *) TOP);
   DROP;
} 
//*****************************************************************************
void N_Class_classMethod(register CThread *thread)
// VALIDATED *************************************
{
MyMessage *m;

   m = (MyMessage *) SECOND; 
   checkType(thread, m, Message);

   SECOND = (Cell) findClassMethod(m, (Class *) TOP);
   DROP;
} 
//*****************************************************************************
void N_Class_is(register CThread *thread)
//***************************************
{
Class *cl;
MyProperty *p;
MyNodeProp *np;
Attribute *na;

   cl = (Class *) SECOND;
   p  = (MyProperty *) TOP;
   checkType(thread, p, Property);
   
   np = (MyNodeProp *) allocDictCell(thread, sizeof(MyNodeProp) / sizeof(Cell));
   np->next = cl->propList;
   np->prop = p;
   cl->propList = np;

   for(na =  p->attList; na != NULL; na = na->next)
      add_attribute(thread, cl, na->name, p, na->isMutable);
      
   DROP2;
}
//*****************************************************************************
void N_Constant_initialize(register CThread *thread)
//**************************************************
{
Cell value;

   value = (Cell) SECOND;
   checkImmutable(thread, (MyObject *) value);
   setObjPersistent(thread, (MyObject *) value);
   ((MyConstant *) TOP)->value = value;
   DROP2;
}
//*****************************************************************************
void N_Variable_initialize(register CThread *thread)
// VALIDATED ***************************************
{
MyVariable *v;

   POPDS(v, MyVariable *); 
   v->index = LockInc(thread->vm->nbVars);
}
//*****************************************************************************
void N_Package_initialize(register CThread *thread)
//*************************************************
{
MyPackage *pack;

   POPDS(pack, MyPackage *); 
   pack->wrdlist = allocDictArray(thread, HASH_SIZE);
}
//*****************************************************************************
void N_Package_import(register CThread *thread)
//*********************************************
{
MyPackage *pack, *cur;
MyImportedPack *ip;

   pack = (MyPackage *) TOP;
   cur  = currentPackage(thread);
   if (cur == NULL) cur = rootPackage(thread);

   ip = (MyImportedPack *) allocDictCell(thread, sizeof(MyImportedPack) / sizeof(Cell));
   ip->next = cur->importedPackages;
   ip->package = pack;
   cur->importedPackages = ip;
   DROP;
}
//*****************************************************************************
void N_Function_initialize(CThread *thread)
//*****************************************
{
Func *f;

   POPDS(f, Func *);
   f->code = (fptr) targetCode(thread);
   initCode(thread);
}
//*****************************************************************************
void N_Dual_defCompile(CThread *thread)
//*************************************
{
FDual *d;

   POPDS(d, FDual *);
   if (d->compileCode != NULL)
      sysExceptionFromC(thread, "Can't redefine compile action");

   d->compileCode = (fptr) targetCode(thread);
   initCode(thread);
   setvar(thread, STATE, (Cell) OONE);
}
//*****************************************************************************
void N_implement(CThread *thread)
//*******************************
{
MyMessage *s;
MyMethod *meth;

   s = (MyMessage *) TOP;
   checkType(thread, s, Message);
   meth = add_meth(thread, s, (Class *) THIRD, (Class *) FOURTH, NULL, (Class *) SECOND);
   meth->code = (fptr) targetCode(thread);
   initCode(thread);
}
//*****************************************************************************
void N_colend(CThread *thread)
//****************************
{
MyTask *t;
CodePage *page;
uchar *cp;
vint codeSize;

   N_RET(thread);
   t = thread->currentTask;
   page = t->codePage;

   // End of def and save current definition into target code page
   cp = (uchar *) t->codeBuffer + indexCode(thread);
   do { *cp = '\0'; cp++; } while (((vint) cp) % sizeof(Cell) != 0);

   codeSize = cp - t->codeBuffer;
   // dbg("Code = %x\n", page->target);

   if (PageProtectExW(page) == PageProtError) compileError();
   memcpy(page->target, t->codeBuffer, codeSize);
   page->target += codeSize;
   if (PageProtectEx(page)  == PageProtError) compileError();

   setvar(thread, STATE, (Cell) OZERO);
}
//*****************************************************************************
void N_Block_new(CThread *thread)
//*******************************
{
MyBlock *bl;

   bl = (MyBlock *) allocDictObject(thread, Block, sizeof(MyBlock));
   bl->code = (fptr) (targetCode(thread) + indexCode(thread));
   bl->nbVars = bl->varused = 0;
   PUSHDS(bl);
}
//*****************************************************************************
void initClosure(CThread *thread)
//*******************************
{
MyClosure *closure;
int var, i;
vint msk, objp, blmsk;
MyObject **vars, *framevar;
MyBlock *bl;
PCell frame;

   closure = (MyClosure *) THIRD;
   frame = (PCell) SECOND;
   bl = (MyBlock *) TOP;

   var = bl->nbVars; 
   closure->objptr   = (bl->objptr & MSKCLASSID & MSKMODERUN) 
                       + 2 * (MSKATTHIDDEN + MSKATT) + var * MSKATT;
   closure->code     = bl->code;
   closure->nbVars   = var;

   vars = &(closure->vars);
   blmsk = bl->varused;
   msk = 2;

   if (bl->objptr & MSKMUTABLE) closure->objptr |= MSKMUTABLE ;

   for(i = 0; i < var; i++) { 
      if (blmsk & msk) { 
         vars[i] = framevar = (MyObject *) *frame;
         if (!ISINTNIL(framevar)) {
            objp = framevar->objptr;
            if (!(objp & MSKPERSISTENT) && (objp & MSKMUTABLE))
               closure->objptr |= MSKMUTABLE;
            }
         }
      else { vars[i] = (MyObject *) nil; }
      frame++;
      msk = msk << 1;
      }

    DROP2;
}
//*****************************************************************************


//*****************************************************************************
// ARRAY METHODS
//*****************************************************************************

//*****************************************************************************
void arrayInit(register CThread *thread)
//**************************************
{
MyArray *l;
register vint size;

   l = (MyArray *) TOP;
   size = (vint) SECOND;
   checkTypeInteger(thread, size);

   l->buf = l->free = allocMem(thread, TOCINT(size) * sizeof(Cell), (char **) &(l->alloc));
   DROP2
}
//*******************************************************************************
void N_Array_realloc(register CThread *thread)
// VALIDATED *********************************
{
register MyArray *a;
vint sizeUsed;
Cell *newBuf;

   a = (MyArray *) TOP;
   sizeUsed = a->free - a->buf;
   newBuf = allocMem(thread, sizeUsed * 2 * sizeof(Cell), (char **) &(a->alloc));
   memcpy(newBuf, a->buf, sizeUsed * sizeof(Cell) );
   freeMem(thread, (PTR) a->buf);
   a->buf = newBuf;
   a->free = a->buf + sizeUsed;
}
//*******************************************************************************
void N_Array_removeAt(register CThread *thread)
// VALIDATED **********************************
{
register Cell *c, *free;
register MyArray *a;
vint i;
MyObject *o;

   CHECKSTACK(2)

   a = (MyArray *) TOP; i = TOCINT(SECOND) - 1; DROP2;
   if (i >= (vint) (a->free - a->buf)) { PUSHDS(nil); return; }

   c = &(a->buf[i]); o = (MyObject *) *c; PUSHDS(o); 

   // If during GC, check if a reference is not crushed
   if (thread->gcPhase > GC_SWEEPING) pushMarkStack(thread, o);

   // All elements after this one have their index decreased
   free = a->free - 1; 
   while (c != free) { *c = *(c+1); c++; }
   a->free--; 
}
//*******************************************************************************


//*******************************************************************************
// BUFFER
//*******************************************************************************

#define JUSTIFY_LEFT       0
#define JUSTIFY_RIGHT      1

//*******************************************************************************
void bufferInit(register CThread *thread)
//***************************************
{
MyBuffer *b;
register vint size;

   b = (MyBuffer *) TOP;
   size = (vint) SECOND;
   checkTypeInteger(thread, size);

   size = (vint) TOCINT(size);
   b->buf = allocMem(thread, size + 1, NULL);
   b->free1 = ((MyBuffer *) b)->buf + size;
   DROP2;
}
//*******************************************************************************
void bufferFinalize(register CThread *thread, MyObject *b)
//********************************************************
{
register uchar *buf;

   buf = ((MyBuffer *) b)->buf;
   if (buf != nil) freeMem(thread, (PCell) buf);
}
//*******************************************************************************
void N_Buffer_utf8At(CThread *thread)
//***********************************
{
MyString *s;
MYINT oindex, iindex, index;
register uchar *buf;
register int c;
register uchar d;
register vint size, translate, i;

   POPDS(s, MyString *); buf = s->buf;
   POPDS(oindex, MYINT); iindex = TOCINT(oindex);
   index = iindex - 1;
   if (index < 0 || index >= s->free1 - buf) { PUSHDS(nil); return; }
   c = buf[index];
   if (c <= 0x7F) { PUSHDS(TOOINT(iindex+1)); PUSHDS(TOOINT(c)); return; }

   if (c >= 0xC2 && c <= 0xDF)        { size = 1; translate = 0x0003080; }
   else { if (c >= 0xE0 && c <= 0xEF) { size = 2; translate = 0x00E2080; }
   else { if (c >= 0xF0 && c <= 0xF4) { size = 3; translate = 0x3C82008; }
   else { PUSHDS(nil); return ; } } }

   for(i = 1; i <= size; i++) {
      d = buf[index + i ];
      if (d < 0x80 || d > 0xBF) { PUSHDS(nil); return; }
      if (i == 1) {
         if (size == 2) {
            if (c == 0xE0 && d < 0xA0) { PUSHDS(nil); return ; }
            if (c == 0xED && d > 0x9F) { PUSHDS(nil); return ; }
            }
         if (size == 3) {
            if (c == 0xF0 && d < 0x90) { PUSHDS(nil); return; }
            if (c == 0xF4 && d > 0x8F) { PUSHDS(nil); return; }
            }
         }
      c = (c << 6) + d;
      }
   c -= translate;
   PUSHDS(TOOINT(iindex + size + 1)); PUSHDS(TOOINT(c));
}
//*******************************************************************************
void gen_format(char *format, char type, vint width, vint justify, vint prec)
//***************************************************************************
{
vint i = 1;
   format[0] = '%';

   // Justify
   if (justify == JUSTIFY_LEFT) { format[i++] = '-'; } 
   
   // width
   if (width > 0 && width <= 30) {
      if (width >= 10) format[i++] = '0' + (width / 10);
      format[i++] = '0' + (width % 10);
      }

   // precision
   if (prec > 0 && prec <= 30) {
      format[i++] = '.';
      if (prec >= 10) format[i++] = '0' + (prec / 10);
      format[i++] = '0' + (prec % 10); 
      }

   format[i++] = type;
   format[i] = '\0';
   // dbg("format = %s\n", format);
}
//*******************************************************************************
void N_Buffer_hashValue(register CThread *thread)
// VALIDATED ************************************
{
   TOP = (Cell) TOOINT((hash_value((char *) ((MyString *) TOP)->buf))); 
}
//*******************************************************************************
void N_String_asSymbol(register CThread *thread)
//**********************************************
{
   TOP = (Cell) add_symbol(thread, (char *) ((MyString *) TOP)->buf);
}
//*******************************************************************************


//*******************************************************************************
// STRING METHODS
//*******************************************************************************

//*****************************************************************************
vint utf8ToChars(register CThread *thread, register vint c)
//*********************************************************
{
   if (c < 0x80) { FORMAT[0] = c; FORMAT[1] = '\0'; return 1; }

   if (c < 0x800) {
      FORMAT[0] = (c >> 6) | 0xC0;
      FORMAT[1] = (c & 0x3F) | 0x80;
      FORMAT[2] = '\0';
      return 2;
      }
            
   if (c < 0x10000) {
      FORMAT[0] = (c >> 12) | 0xE0;
      FORMAT[1] = (c >> 6 & 0x3F) | 0x80;
      FORMAT[2] = (c & 0x3F) | 0x80;
      FORMAT[3] = '\0';
      return 3;
      }

   if (c < 0x110000) {
      FORMAT[0] = (c >> 18) | 0xF0;
      FORMAT[1] = (c >> 12 & 0x3F) | 0x80;
      FORMAT[2] = (c >> 6 & 0x3F) | 0x80;
      FORMAT[3] = (c & 0x3F) | 0x80;
      FORMAT[4] = '\0';
      return 4;
   }

   sysExceptionFromC(thread, "Bad UTF8 encoding");
   return 0;
}
//*****************************************************************************
void bufAddChars(CThread *thread, register MyBuffer *b, uchar *toadd, vint sizetoadd)
//*******************************************************************************
{
vint sizeprev;
uchar *newBuf;

   if (b->alloc - b->free1 <= sizetoadd) { 
      sizeprev = b->free1 - b->buf;
      newBuf = allocMem(thread, (sizeprev + sizetoadd) * 2, (char **) &(b->alloc));
      memcpy(newBuf, b->buf, sizeprev);
      freeMem(thread, (PTR) b->buf);
      b->free1 = newBuf + sizeprev;
      // dbg("Realloc %x to %x, free=%x alloc=%x\n", b->buf,newBuf,b->free1,b->alloc);
      b->buf = newBuf;
      }

   if (toadd != NULL) {
      memcpy(b->free1, toadd, sizetoadd + 1);
      b->free1 += sizetoadd; 
      }
}
//*****************************************************************************
void stringInit(register CThread *thread)
//***************************************
{
MyString *sb;
register vint size;

   sb = (MyString *) TOP;
   size = (vint) SECOND;
   checkTypeInteger(thread, size);

   size = (vint) TOCINT(SECOND) + 1;
   sb->buf = sb->free1 = allocMem(thread, size, (char **) &(sb->alloc));
   *(sb->buf) = '\0';
   DROP2;
}
//*****************************************************************************
void N_String_add(register CThread *thread)
//*****************************************
{
register MyString *s;
register vint c;

   s = (MyString *) TOP;
   c = (vint) TOCINT(SECOND); 
   bufAddChars(thread, (MyBuffer *) s, FORMAT, utf8ToChars(thread, c));
   DROP2;
}
//*****************************************************************************
void N_String_addSymbol(register CThread *thread)
//***********************************************
{
register MyString *s;
register MySymbol *sym;
   
   CHECKSTACK(2)

   s = (MyString *) TOP;
   sym = (MySymbol *) SECOND;
   checkType(thread, sym, Symbol);

   bufAddChars(thread, (MyBuffer *) s, (uchar *) symbuf(sym), strlen(symbuf(sym)));
   DROP2;
}
//*****************************************************************************
void N_String_addBuffer(register CThread *thread)
//***********************************************
{
register MyString *str;
register MyBuffer *source;
   
   CHECKSTACK(2)

   str       = (MyString *) TOP;
   source    = (MyBuffer *) SECOND;
   checkType(thread, source, Buffer);

   bufAddChars(thread, (MyBuffer *) str, source->buf, source->free1 - source->buf);
   DROP2;
}
//*****************************************************************************
void N_String_addIntegerFormat(register CThread *thread)
//******************************************************
{
char format[16] = {0};
register MyString *str;
MYINT sourceInt, w, j, p;
MyNatural *n;
int size;
vint *res;
   
   CHECKSTACK(5)

   str       = (MyString *) TOP;
   sourceInt = (sint) FIFTH;
   checkType(thread, (void *) sourceInt, Integer);

   w = TOCINT(FOURTH);
   j = TOCINT(THIRD);
   p = TOCINT(SECOND);

   if (ISINT(sourceInt)) {
      gen_format(format, 'd', w, j, p);
      sprintf((char *) FORMAT, format, TOCINT(sourceInt));
      bufAddChars(thread, (MyBuffer *) str, FORMAT, strlen((char *) FORMAT));
      }
   else {
      n = (MyNatural *) sourceInt;
      if (n->sign == -1) bufAddChars(thread, (MyBuffer *) str, (uchar *) "-", 1);
   
      res = (vint *) allocMem(thread, (n->free - n->buf) * 2 * sizeof(vint), NULL);
      decimalValues(thread, n, res, &size);

      sprintf((char *) FORMAT, "%d", res[size]);
      bufAddChars(thread, (MyBuffer *) str, FORMAT, strlen((char *) FORMAT));
      for(size-- ; size >= 0 ; size--) {
         sprintf((char *) FORMAT, "%.9d", res[size]);
         bufAddChars(thread, (MyBuffer *) str, FORMAT, strlen((char *) FORMAT));
         }

      freeMem(thread, (PTR) res);
      }

   DROP5;
}

#ifdef __FLOAT_ENABLE__
//*****************************************************************************
void N_String_addFloatFormat(register CThread *thread)
//****************************************************
{
char format[16] = {0};
register MyString *str;
MyFloat *source;
   
   CHECKSTACK(5)

   str       = (MyString *) TOP;

   source    = (MyFloat *) FIFTH;
   checkType(thread, source, Float);

   gen_format(format, 'g', TOCINT(FOURTH), TOCINT(THIRD), TOCINT(SECOND));
   sprintf((char *) FORMAT, format, source->f);
   bufAddChars(thread, (MyBuffer *) str, FORMAT, strlen((char *) FORMAT));
   DROP5;
}
#endif

//*****************************************************************************
void N_strAsInteger(register CThread *thread)
//*******************************************
{
uchar *str, *ends;
MYINT n;

   str = ((MyString *) TOP)->buf;
   n = strtol((char *) str, (char **) &ends, 0);
   if (n < MIN_CINT || n > MAX_CINT || *ends != 0) { TOP = (Cell) nil; }
   else { TOP = (Cell) TOOINT(n); }
}
//*****************************************************************************
void N_String_asFloat(register CThread *thread)
//*********************************************
{
uchar *str, *ends;
MyFloat *mf;

   str = ((MyString *) SECOND)->buf;
   mf = (MyFloat *) TOP;
   mf->f = strtod((char *) str, (char **) &ends);
   if (ends != str + strlen((char *) str)) { SECOND = (Cell) nil; }
   else { SECOND = (Cell) mf; }
   DROP;
}
//*****************************************************************************


//*****************************************************************************
// FILE METHODS
//*****************************************************************************

#define FILEBEGIN    0
#define FILECURRENT  1
#define FILEEND      2

char *FileAccess[3] = { "r", "w", "a" };

//*****************************************************************************
void N_fileNotFound(CThread *thread)
//**********************************
{
MyString *name;
   POPDS(name, MyString *);
   sysExParamFromC(thread, "Can't open source file", name);
}
//*****************************************************************************
void cantRead(register CThread *thread, MyFile *f)
// VALIDATED *************************************
{
   sysExParamFromC(thread, "Can't read file", f);
}
//*****************************************************************************
void cantWrite(CThread *thread, MyFile *f)
// VALIDATED *****************************
{
   sysExParamFromC(thread, "Can't write on file", f);
}
//*****************************************************************************
void notOpen(CThread *thread, MyFile *f)
// VALIDATED ***************************
{
   sysExParamFromC(thread, "File is not open", f);
}
//*****************************************************************************
vint fileClose(register MyFile *f)
// VALIDATED *********************
{
int ret = 0;

   if (!ISNIL(f->h)) { ret = fclose(f->h); f->h = (FILE *) nil; }
   return ret;
}
//*****************************************************************************
void fileFinalize(register CThread *thread, MyObject *f)
//******************************************************
{
   fileClose((MyFile *) f);
}
//*****************************************************************************
void N_File_open(register CThread *thread)
// VALIDATED *****************************
{
register MyFile *of;
vint access;
char mode[4] = {0};
FILE *f;

   CHECKSTACK(2)

   of = (MyFile *) TOP; 

   access = TOCINT(SECOND);
   if (access > 2) sysExceptionFromC(thread, "Bad mode to open file");

   strcpy(mode, FileAccess[access]);
   if (of->encoding == FILE_BIN) strcat(mode, "b");

   if ((f = fopen((char *) of->name->buf, mode)) == NULL) sysExOS(thread);

   of->h = f;
   DROP2;
}
//*****************************************************************************
long N_File_position(register CThread *thread)
//********************************************
{
register MyFile *f;
long int pos;

   f = (MyFile *) TOP; 
   pos = ftell(f->h);
   if (pos == -1) sysExOS(thread);
   return pos;
}
//*******************************************************************
void N_File_reposition(register CThread *thread)
//**********************************************
{
register MyFile *f;
long int offset;
int origin;
MYINT opos;

   CHECKSTACK(3)

   f = (MyFile *) TOP; 

   origin = (int) TOCINT(THIRD);
   if (origin < FILEBEGIN && origin > FILEEND) 
      sysExceptionFromC(thread, "Bad origin value into #reposition");

   if (origin == FILEBEGIN) origin = SEEK_SET; 
   else if(origin == FILECURRENT) origin = SEEK_CUR;
   else origin = SEEK_END;

   opos = (MYINT) SECOND;
   checkTypeInteger(thread, opos);

   offset = TOCINT(opos);
   if (fseek(f->h, offset, origin) != 0) sysExOS(thread);

   DROP3;
}
//*****************************************************************************
void N_File_close(register CThread *thread)
// VALIDATED ******************************
{
register MyFile *f;

   f = (MyFile *) TOP; 
   if (fileClose(f) != 0) sysExOS(thread);
   DROP; 
}
//*****************************************************************************
void N_File_stats(register CThread *thread)
//*****************************************
{
MyString *name;
MYLONG *l;
char *ds;
    
#ifdef __OWIN__
struct _stati64 res;
#endif
#ifdef __OLNX__
struct stat res;
#endif

    POPDS(name, MyString *);
    checkType(thread, name, String);

    if (
#ifdef __OWIN__
      _stati64((char *) name->buf, &res) != 0
#endif
       
#ifdef __OLNX__
      stat((char *)     name->buf, &res) != 0
#endif
      )
      { PUSHDS(nil); PUSHDS(nil); PUSHDS(nil); PUSHDS(OTRUE); }
   else {
      PUSHDS(OFALSE);
      ds = (char *) (DS - 2);
      ds -= sizeof(MYLONG); l = (MYLONG *) ds; *l = ((MYLONG) res.st_mtime) * 1000000;
      ds -= sizeof(MYLONG); l = (MYLONG *) ds; *l = ((MYLONG) res.st_ctime) * 1000000;
      ds -= sizeof(MYLONG); l = (MYLONG *) ds; *l = (MYLONG)  res.st_size;
      }
}
//*****************************************************************************
void N_File_addChar(register CThread *thread)
// VALIDATED ********************************
{
register MyFile *f;
register vint c;

   CHECKSTACK(2)

   f = (MyFile *) TOP; 
   if (ISNIL(f->h)) sysExOS(thread);

   c = (vint) SECOND; 
   checkTypeInteger(thread, c);

   c = (vint) TOCINT(c);

   if (f->encoding == FILE_UTF8 && c >= 0x80) {
      utf8ToChars(thread, c);
      if (fprintf(f->h, (char *) FORMAT) < 0) sysExOS(thread); 
      }
   else { if (fputc(c, f->h) == -1) sysExOS(thread); }

   DROP2;
}
//*****************************************************************************
Cell N_File_addSymbol(register CThread *thread)
//********************************************
{
MyFile *f;
register MySymbol *s;
   
   CHECKSTACK(2)

   f = (MyFile *) TOP;
   s = (MySymbol *) SECOND;
   checkType(thread, s, Symbol);

   if (fprintf(f->h, "%s", symbuf(s)) < 0) sysExOS(thread);

   DROP2; return NULL;
}
//*****************************************************************************
void N_File_addBufferFormat(register CThread *thread)
// VALIDATED **************************************
{
MyFile *f;
register MyBuffer *source;
vint len;
   
   CHECKSTACK(2)

   f = (MyFile *) TOP;

   source    = (MyBuffer *) SECOND;
   checkType(thread, source, Buffer);
 
   len = source->free1 - source->buf;
   if ( ISNIL(f->h) || fwrite((char *) source->buf, 1, len, f->h) < 0 )
       { sysExOS(thread); }

   DROP2;
}
//*****************************************************************************
void N_File_addIntegerFormat(register CThread *thread)
//****************************************************
{
register MyFile *f;
MYINT sourceInt, w, j, p;
MyNatural *n;
int size;
vint *res;
    
   CHECKSTACK(5)

   f = (MyFile *) TOP;
   if (ISNIL(f->h)) cantWrite(thread, f);

   sourceInt = (sint) FIFTH;
   checkType(thread, (void *) sourceInt, Integer);

   w = TOCINT(FOURTH);
   j = TOCINT(THIRD);
   p = TOCINT(SECOND);

   if (ISINT(sourceInt)) {
      gen_format((char *) FORMAT, 'd', w, j, p);
      if (fprintf(f->h, (char *) FORMAT, TOCINT(sourceInt)) < 0) sysExOS(thread);
      }
   else {
      n = (MyNatural *) sourceInt;
      if (n->sign == -1 && fputc('-', f->h) == EOF) sysExOS(thread);
   
      res = (vint *) allocMem(thread, (n->free - n->buf) * 2 * sizeof(vint), NULL);
      decimalValues(thread, n, res, &size);

      if (fprintf(f->h, "%d", res[size]) < 0) sysExOS(thread);
      for(size--; size >= 0; size--) 
         { if (fprintf(f->h, "%.9d", res[size]) < 0) sysExOS(thread); }

      freeMem(thread, (PTR) res);
      }
   
   DROP5; 
}

#ifdef __FLOAT_ENABLE__
//*****************************************************************************
void N_File_addFloatFormat(register CThread *thread)
//**************************************************
{
register MyFile *f;
MyFloat *source;
   
   CHECKSTACK(5)

   f = (MyFile *) TOP; if (ISNIL(f->h)) sysExOS(thread);

   source    = (MyFloat *) FIFTH;
   checkType(thread, source, Float);

   gen_format((char *) FORMAT, 'g', TOCINT(FOURTH), TOCINT(THIRD), TOCINT(SECOND));
   if (fprintf(f->h, (char *) FORMAT, source->f) < 0) sysExOS(thread);

   DROP5;
}
#endif

//*****************************************************************************
void N_File_flush(register CThread *thread)
//*****************************************
{
register MyFile *f;

   POPDS(f, MyFile *);
   if (f->h == (FILE *) nil) cantWrite(thread, f);
   if (fflush(f->h) != 0) sysExOS(thread);
}
//*****************************************************************************
Cell N_File_atEnd(register CThread *thread)
// VALIDATED ******************************
{
MyFile *f;

   f = (MyFile *) TOP;
   if (f->h == (FILE *) nil) cantRead(thread, f);
   if (feof(f->h)) { TOP = (Cell) OTRUE; } else { TOP = (Cell) OFALSE; }
   return NULL;
}
//*****************************************************************************
void N_File_getChar(register CThread *thread)
// VALIDATED ********************************
{
MyFile *f;
FILE *h;
unsigned char d;
vint c, size, translate, i;

   f = (MyFile *) TOP; h = f->h; 
   if (ISNIL(h)) notOpen(thread, f);

   c = fgetc(h);

   if (f->encoding == FILE_UTF8 && c > 0x7F && c != (uchar) EOF) {
      // dbg("c = %x eof = %x\n", c, (unsigned char) EOF);
      if (c >= 0xC2 && c <= 0xDF)        { size = 1; translate = 0x0003080; }
      else { if (c >= 0xE0 && c <= 0xEF) { size = 2; translate = 0x00E2080; }
      else { if (c >= 0xF0 && c <= 0xF4) { size = 3; translate = 0x3C82008; }
      else { goto LUTF8Error; } }}

      for(i = 1; i <= size; i++) {
         d = fgetc(h);
         if (d < 0x80 || d > 0xBF) goto LUTF8Error;
         if (i == 1) {
            if (size == 2) {
            if (c == 0xE0 && d < 0xA0) goto LUTF8Error;
            if (c == 0xED && d > 0x9F) goto LUTF8Error;
               }
            if (size == 3) {
            if (c == 0xF0 && d < 0x90) goto LUTF8Error;
            if (c == 0xF4 && d > 0x8F) goto LUTF8Error;
               }
            }
         c = (c << 6) + d;
         }
      c -= translate;
      }

LFileGetReturn: 
   if (c == -1) { if (!feof(h)) cantRead(thread, f); c = nil; }
    else { c = TOOINT(c); }
   TOP = (Cell) c;
   return;

LUTF8Error:
   c = '?'; goto LFileGetReturn;
}
//*****************************************************************************
void N_File_readWith(register CThread *thread)
//********************************************
{
MyBuffer *m;
MYINT sizeToRead, size;
int ret;
MyFile *f;
FILE *h;

   CHECKSTACK(3)

   f = (MyFile *) TOP; h = f->h;
   if (ISNIL(h)) notOpen(thread, f);

   size = (MYINT) THIRD;
   checkTypeInteger(thread, size);

   sizeToRead = TOCINT(size);
   if (sizeToRead <= 0) badParams(thread);

   m = (MyBuffer *) SECOND; 
   checkType(thread, m, Buffer);

   if ((m->objptr & MSKMUTABLE) == 0) 
      sysExceptionFromC(thread, "Immutable rule violation");

   if (m->free1 - m->buf < sizeToRead) bufAddChars(thread, m, NULL, sizeToRead);

   ret = fread(m->buf, 1, sizeToRead, h);
   if (ret != sizeToRead && !feof(h)) cantRead(thread, f);
   m->free1 = m->buf + ret;
   THIRD = (Cell) m; DROP2;
}
//*****************************************************************************
void N_File_readLineWith(register CThread *thread)
//************************************************
{
#define LINESIZE     240
MyFile *f;
FILE *h;
MyString *str;
uchar *s = NULL, c;
vint size;
int end = KO;

   POPDS(f, MyFile *); h = f->h; 
   if (ISNIL(h)) notOpen(thread, f);

   str = (MyString *) TOP;
   if ((str->objptr & MSKMUTABLE) == 0) 
      sysExceptionFromC(thread, "Immutable rule violation");

   if (f->encoding != FILE_UTF8 && isKindOf((MyObject *) str, Buffer) == KO)
      sysExceptionFromC(thread, "Buffer required into #readLine");

   str->free1 = str->buf; *(str->buf) = '\0';

   while(1) {
      bufAddChars(thread, (MyBuffer *) str, NULL, LINESIZE);
      s = (uchar *) fgets((char *) str->free1, LINESIZE, h); 

      if (ferror(h)) { 
         str->free1 = str->buf; *(str->buf) = '\0';
         sysExOS(thread);
         }

      if (s == NULL) break;

      size = strlen((char *) s);
      s += size; c = *(s-1);
      while (c == '\n' || c == '\r') { end = OK; s--; c = *(s-1); }
      *s = '\0';
      str->free1 = s;
      if (feof(h) || end == OK) break;
      }
         
   if (s == NULL && str->buf == str->free1)
   TOP = (Cell) nil;
}
//****************************************************************************
FILE *openSource(char *nd)
//************************
{
FILE *f;
char *s;

   s = nd;
   while (*s) { 
      if (*s == '\\') *s = '/';
      if (*s == '.' && *(s+1) == '.') { return NULL; }
      s++;
      }

   // dbg("Trying raw : %s\n", nd);
   f = fopen(nd, "r"); 
   return f;
}
//*****************************************************************************
void findAndOpenFile(CThread *thread)
// VALIDATED ************************
{
MyString *s;
char *name;
char nd[512];
FILE *f;
vint i, nPath;

   POPDS(s, MyString *); name = (char *) s->buf;
   sprintf(nd, "%s", name);
   // dbg("Trying : %s\n", nd);
   f = openSource(nd); 
   if (f != NULL) { PUSHDS(FTOOINT(f)); return ; }

   sprintf(nd, "./%s", name);
   // dbg("Trying : %s\n", nd);
   f = openSource(nd); 
   if (f != NULL) { PUSHDS(FTOOINT(f)); return ; }

   nPath = allPath->free - allPath->buf;
   for(i = 0; i < nPath; i++) {
      sprintf(nd, "%s/%s", ((MyString *) allPath->buf[i])->buf, name);
      // dbg("Trying : %s\n", nd);
      f = openSource(nd);
      if (f != NULL) { PUSHDS(FTOOINT(f)); return ; }
      }

   PUSHDS(nil);
}
//*****************************************************************************


//*****************************************************************************
// Integer METHODS
//*****************************************************************************

//*****************************************************************************
void integerInit(register CThread *thread)
//****************************************
{
register MyNatural *n;
register MYINT size;

   n = (MyNatural *) TOP;
   size = (MYINT) TOCINT(SECOND);
   n->buf = (vint *) allocMem(thread, size * sizeof(Cell), NULL);
   n->free = n->buf + size;
   n->sign = 0;
   DROP2;
}
//*****************************************************************************
void I_longResult(register CThread *thread)
//*****************************************
{
register MyNatural *res;
register MYLONG *l;
register vint i = 0;

   res = (MyNatural *) TOP;
   l = (MYLONG *) (((MyNatural *) SECOND)->buf);
   // dbg("long = %I64d\n", *l);
   if (*l < 0) { res->sign = -1; *l = -(*l); } else { res->sign = 1; } 

   while(*l != 0) {
      res->buf[i] = (vint) ((*l) & NATURAL_LOW);
      // dbg("res[%d] = %d %x\n", i, res->buf[i], res->buf[i]);
      *l >>= NATURAL_SHIFT;
      i++;
      }
   res->free = res->buf + i;

   SECOND = (Cell) res; DROP;
}
//*****************************************************************************
void N_Integer_eq(register CThread *thread)
//*****************************************
{
MyNatural *n, *m;
vint nb, k;

   n = (MyNatural *) TOP;
   m = (MyNatural *) SECOND;

   if (n->sign != m->sign) goto LIntegerFalse;

   nb = n->free - n->buf;
   if (nb != m->free - m->buf) goto LIntegerFalse;

   for(k = 0; k < nb; k++) 
      if (n->buf[k] != m->buf[k]) goto LIntegerFalse;

   SECOND = (Cell) OTRUE; DROP; return;

LIntegerFalse:
   SECOND = (Cell) OFALSE; DROP;
}
//*****************************************************************************
int naturalCompare(MyNatural *n, MyNatural *m, int negm)
//******************************************************
{
int t = 1, f = -1;
vint *vn, *vm;
int sgnm;

   sgnm = m->sign * negm;

   if (n->sign < sgnm) return 1;
   if (n->sign > sgnm) return -1;

   if (n->sign == -1) { t = -1; f = 1; } 

   if (n->free - n->buf < m->free - m->buf) goto LNaturalCmpTrue;
   if (n->free - n->buf > m->free - m->buf) goto LNaturalCmpFalse;

   vm = m->free - 1;

   for(vn = n->free - 1; vn >= n->buf; vn--) {
      if (*vn < *vm) goto LNaturalCmpTrue;
      if (*vn > *vm) goto LNaturalCmpFalse;
      vm--;
      }
   return 0;

LNaturalCmpTrue:
   return t; 
LNaturalCmpFalse:
   return f; 
}
//*****************************************************************************
void N_Integer_cmp(register CThread *thread)
//******************************************
{
MyNatural *n, *m;

   n = (MyNatural *) SECOND;
   m = (MyNatural *) TOP;

   if (naturalCompare(n, m, 1) == -1) SECOND = (Cell) OFALSE;
   else SECOND = (Cell) OTRUE; 
   DROP;
}
//*****************************************************************************
MyNatural *normalize(MyNatural *n)
//********************************
{
   while(*(n->free - 1) == 0 && n->free != n->buf) { n->free--; }

   if (n->free - n->buf > 1) return n;
   return (MyNatural *) TOOINT(n->buf[0] * n->sign);
}
//*****************************************************************************
MyNatural *myNaturalAdd(CThread *thread, MyNatural *n, MyNatural *m, MyNatural *res)
//**********************************************************************************
{
register vint r = 0, i;
vint nsize, msize, max;

   nsize = n->free - n->buf;
   msize = m->free - m->buf;
   max = res->free - res->buf;
   res->sign = n->sign;

   for(i = 0; i < max; i++) {
      if (i < nsize) r += n->buf[i]; 
      if (i < msize) r += m->buf[i]; 
      res->buf[i] = (vint) (r & NATURAL_LOW);
      r >>= NATURAL_SHIFT;
      }
   return res;
}
//*****************************************************************************
MyNatural *myNaturalSub(CThread *thread, MyNatural *n, MyNatural *m, MyNatural *res, int negm)
//****************************************************************************
{
MyNatural *tmp;
vint nsize, msize, i, max;
int q, v, s;

   s = n->sign;
   if (naturalCompare(n, m, negm) != -1) 
      { if (s == 1) { tmp = n; n = m; m = tmp; s = -1; } } 
   else { if (s == -1) { tmp = n; n = m; m = tmp; s = 1; } }

   nsize = n->free - n->buf;
   msize = m->free - m->buf;

   if (nsize >= msize) { max = nsize + 1; } else { max = msize + 1; }

   res->sign = s;

   q = 0;
   for(i = 0; i < max; i++) {
      if (i < nsize) { v = n->buf[i]; } else { v = 0; }
      if (i < msize) { v -= m->buf[i]; }
      v -= q;
      if (v < 0) { v += NATURAL_BASE; q = 1; } else { q = 0; }
      res->buf[i] = v;
      }
   return res;
}
//*****************************************************************************
void N_Integer_add(register CThread *thread)
//******************************************
{
MyNatural *l1, *l2, *res;

   l1 = (MyNatural *) THIRD;
   l2 = (MyNatural *) SECOND;
   res = (MyNatural *) TOP;

   if (l1->sign == l2->sign) res = myNaturalAdd(thread, l1, l2, res);
   else                      res = myNaturalSub(thread, l1, l2, res, -1);

   THIRD = (Cell) normalize(res); DROP2
}
//*****************************************************************************
void N_Integer_sub(register CThread *thread)
//******************************************
{
MyNatural *l1, *l2, *res;

   l1 = (MyNatural *) THIRD;
   l2 = (MyNatural *) SECOND;
   res = (MyNatural *) TOP;

   if (l1->sign != l2->sign) res = myNaturalAdd(thread, l1, l2, res);
   else res = myNaturalSub(thread, l1, l2, res, 1);

   THIRD = (Cell) normalize(res); DROP2;
}
//*****************************************************************************
void N_Integer_mulOne(CThread *thread)
//***********************************
{
MyNatural *l, *res;
MYLONG num;
register vint n;
register MYLONG r;
register vint *lb, *resb;
vint lsize;

   l = (MyNatural *) THIRD; lb = l->buf;
   num = (MYLONG) (TOCINT(SECOND));
   res = (MyNatural *) TOP; resb = res->buf;

   lsize = l->free - lb;

   if (num >= 0) { res->sign = l->sign ; }
   else { res->sign = - l->sign; num = -num; }

   r = 0;
   for(n = 0; n < lsize; n++) {
      r += num * lb[n];
      resb[n] = (vint) (r & NATURAL_LOW);
      r >>= NATURAL_SHIFT;
      }

   if (r == 0) { res->free--; } else { resb[n] = r; }
   THIRD = (Cell) normalize(res); DROP2;
}
//*****************************************************************************
void N_Integer_mul(CThread *thread)
//********************************
{
register vint n, m;
MyNatural *l1, *l2, *res;
vint max, l1size, l2size, v;
MYLONG r;

   l1 = (MyNatural *) THIRD; 
   l2 = (MyNatural *) SECOND;
   res = (MyNatural *) TOP;

   l1size = l1->free - l1->buf;
   l2size = l2->free - l2->buf;
   max = l1size + l2size + 1;
   res->sign = l1->sign * l2->sign;

   for(n = 0; n < max; n ++) res->buf[n] = 0;

   for(n = 0; n < l1size; n++) {
      v = l1->buf[n];
      if (v == 0) continue;
      r = 0;
      for(m = 0; m < l2size; m++) {
         r += ((MYLONG) l2->buf[m]) * ((MYLONG) v) + (MYLONG) res->buf[n+m];
         res->buf[n+m] = (vint) (r & NATURAL_LOW);
         r >>= NATURAL_SHIFT;
         }
      res->buf[n + l2size] = r;
      }

   THIRD = (Cell) normalize(res); 
   DROP2;
}
//*****************************************************************************
void myNaturalRight(vint *target, vint *source, register int len, int shift)
//**************************************************************************
{
MYLONG x = 0, y;

   x = (*source++) >> shift;
   for( len--; len > 0; len--) {
      y = *source++;
      y <<= (NATURAL_SHIFT - shift);
      y |= x;
      *target++ = (vint) (y & NATURAL_LOW);
      x = y >> NATURAL_SHIFT;
      }
   *target = x;
}
//*****************************************************************************
int myNaturalLeft(vint *target, vint *source, register int len, int shift)
//************************************************************************
{
MYLONG x = 0, y;

   for( ;len > 0 ; len--) {
      y = (*source++);
      y <<= shift;
      y |= x;
      *target++ = (vint) (y & NATURAL_LOW);
      x = y >> NATURAL_SHIFT;
      }
   return (int) x;
}
//*****************************************************************************
vint myNaturalDivOne(vint *n, int len, vint divisor, vint *q)
//***********************************************************
{
MYLONG rem, div;

   len--;
   div = (MYLONG) divisor;

   rem = (MYLONG) n[len];
   if (rem >= div) { rem = 0; }
   else { q[len] = 0; len--; rem *= (MYLONG) NATURAL_BASE; }

   for(; len >= 0; len--) {
      rem += (MYLONG) n[len];
      q[len] = rem / div;
      rem = (rem % div) * (MYLONG) NATURAL_BASE;
      }

   return (int) (rem >>= NATURAL_SHIFT);
}
//*****************************************************************************
void decimalValues(CThread *thread, MyNatural *n, vint *res, int *size)
//*********************************************************************
{
vint *quotient;
MYINT len, i;

   len = n->free - n->buf;
   quotient = (vint *) allocMem(thread, len * sizeof(vint), NULL);
   memcpy(quotient, n->buf, len * sizeof(vint));

   i = 0;
   while(len > 0) {
      res[i] = myNaturalDivOne(quotient, len, NATURAL_PRINT, quotient);
      i++;
      while(len > 0 && quotient[len-1] == 0) len--;
      }

   freeMem(thread, (PTR) quotient);

   *size = i - 1; 
}
//*****************************************************************************
/*
void DN(char *s, vint *v, int len)
{
int i;

   dbg("%s : ", s);
   for(i = len - 1; i >= 0; i--) dbg("%.9x | ", v[i]);
   dbg("\n%s : ", s);
   for(i = len - 1; i >= 0; i--) dbg("%.9d | ", v[i]);
   dbg("\n");
}
*/
//*****************************************************************************
void N_Integer_divmod(register CThread *thread)
//*********************************************
{
vint m, n, qlen, rlen, *div, *rem, r;
MyNatural *dividend, *divisor, *quotient, *remainder;
MYLONG w, prod;
int shift, dn, k, l, shiftRequired, qhat, borrow, save, offset, prodl, num;

   dividend = (MyNatural *) FOURTH;
   m = dividend->free - dividend->buf;

   divisor  = (MyNatural *) THIRD;
   n = divisor->free - divisor->buf;

   quotient = (MyNatural *) SECOND;
   qlen = m - n + 1;
   quotient->sign = dividend->sign * divisor->sign;
   
   remainder = (MyNatural *) TOP;
   // divisor if 1 lenght, use myNaturlDivOne
   if (remainder == NULL) {
      FOURTH = (Cell) TOOINT(myNaturalDivOne(dividend->buf, m, 
                                    divisor->buf[0], quotient->buf));
      goto LCanonicQuotient;
      }
      
   rlen = m;
   remainder->sign = dividend->sign;
   remainder->free = remainder->buf + n;

   // Calculate number of bits to shift divisor/dividende and shift them
   div = divisor->buf;
   
   // DN("Divisor:", div, n);

   dn = divisor->buf[n - 1];
   shift = 0;
   for(k = 16; k > 0; k >>= 1) { 
      l = dn >> k;
      if (l == 0) shift += k; else dn = l;
      }
   shift -= 2;
   // dbg("Shift = %d\n", shift);

   if (shift == 0) { shiftRequired = KO; }
   else {
      div = (vint *) allocMem(thread, n * sizeof(vint), NULL);
      myNaturalLeft(div, divisor->buf, n, shift);
      shiftRequired = OK;
      }
   // DN("Divisor", div, n);
   // DN("Dividend: ", dividend->buf, m);

   rem = remainder->buf;
   if (shiftRequired == OK) l = myNaturalLeft(rem, dividend->buf, m, shift);
   else { memcpy(rem, dividend->buf, m * sizeof(vint)); l = 0; }
   rem[rlen] = l;

   // Now, ready to run division
   dn = div[n - 1];

   for(k = rlen; k >= n; k--) {
      if (rem[k] == dn) { qhat = NATURAL_BASE - 1; }
      else { 
         w = (((MYLONG) rem[k]) << NATURAL_SHIFT) + (MYLONG) rem[k-1];
         qhat = (int) (w / (MYLONG) dn);
         }

      if (qhat != 0) {
         // Removes div * qhat from rem[k-n, ... ]
         borrow = 0;
         offset = k - n;

         for(l = 0; l < n; l++) {
            // DN("Remainder01: ", rem, rlen + 1);
            prod   = (MYLONG) div[l] * (MYLONG) qhat + (MYLONG) borrow;
            prodl  = (int) (prod & NATURAL_LOW);
            borrow = (int) (prod >> NATURAL_SHIFT);
            save   = ((int) rem[offset + l]) - prodl;
            if (save < 0) { save += NATURAL_BASE; borrow++; }
            rem[offset+l] = save;
            }

         num = rem[k] - borrow;
         while(num != 0) {
            qhat--;
            r = 0;
            for(l = 0; l < n; l++) {
               r += rem[offset + l] + div[l];
               rem[offset + l] = (vint) (r & NATURAL_LOW);
               r >>= NATURAL_SHIFT;
               }
            rem[k] += r;
            num = r - 1;
            }

         }
      rem[k] = qhat;
      }

   // Calculate quotient & remainder
   for(k = 0; k < qlen; k++) quotient->buf[k] = rem[k + n];

   if (shiftRequired == OK) {
      myNaturalRight(rem, rem, n, shift);
      freeMem(thread, (PTR) div);
      }

   FOURTH = (Cell) normalize(remainder);

LCanonicQuotient:
   THIRD = (Cell) normalize(quotient);
   DROP2;
}
//*****************************************************************************
void N_Integer_bitAnd(register CThread *thread)
//*********************************************
{
MyNatural *l1, *l2, *res;
int max, l1size, l2size;
register int i;

   l1 = (MyNatural *) THIRD;
   l2 = (MyNatural *) SECOND;
   res = (MyNatural *) TOP;

   l1size = l1->free - l1->buf;
   l2size = l2->free - l2->buf;
   max = res->free - res->buf;
   res->sign = 1;

   for(i = 0; i < max; i ++) {
     if (i >= l1size || i >= l2size) { res->buf[i] = 0; }
     else { res->buf[i] = l1->buf[i] & l2->buf[i]; }
     // dbg("i = %d, res = %d, l1 = %d, l2 = %d\n", i, res->buf[i], l1->buf[i], l2->buf[i]);
     }

   THIRD = (Cell) normalize(res);
   DROP2;
}
//*****************************************************************************
void N_Integer_bitOr(register CThread *thread)
//****************************************
{
MyNatural *l1, *l2, *res;
int max, l1size, l2size;
register int i;

   l1 = (MyNatural *) THIRD;
   l2 = (MyNatural *) SECOND;
   res = (MyNatural *) TOP;

   l1size = l1->free - l1->buf;
   l2size = l2->free - l2->buf;
   max = res->free - res->buf;
   res->sign = 1;

   for(i = 0; i < max; i ++) {
       if (i >= l1size) { res->buf[i] = l2->buf[i]; continue; }
       if (i >= l2size) { res->buf[i] = l1->buf[i]; continue; }
       res->buf[i] = l1->buf[i] | l2->buf[i];
       }

   THIRD = (Cell) normalize(res);
   DROP2;
}
//*****************************************************************************
void N_Integer_bitXor(register CThread *thread)
//*********************************************
{
MyNatural *l1, *l2, *res;
int max, l1size, l2size, n1, n2;
register int i;

   l1 = (MyNatural *) THIRD;
   l2 = (MyNatural *) SECOND;
   res = (MyNatural *) TOP;

   l1size = l1->free - l1->buf;
   l2size = l2->free - l2->buf;
   max = res->free - res->buf;
   res->sign = 1;

   for(i = 0; i < max; i ++) {
       if (i >= l1size) { n1 = 0; } else { n1 = l1->buf[i]; }
       if (i >= l2size) { n2 = 0; } else { n2 = l2->buf[i]; }
       res->buf[i] = n1 ^ n2;
       }

   THIRD = (Cell) normalize(res);
   DROP2;
}
//*****************************************************************************
void N_Integer_bitLeft(register CThread *thread)
//*********************************************
{
register MYINT i, left;
MyNatural *n, *res;
int l, cshift, cleft, max;

   res = (MyNatural *) TOP; max = res->free - res->buf;
   n = (MyNatural *) SECOND; l = n->free - n->buf;
   res->sign = n->sign;

   left = (MYINT) TOCINT(THIRD);
   cshift = left / NATURAL_SHIFT;
   cleft  = left % NATURAL_SHIFT;
   
   for(i = 0; i < max; i++) res->buf[i] = 0;

   res->buf[l + cshift] = (vint) myNaturalLeft(res->buf + cshift, n->buf, l, cleft);
   THIRD = (Cell) normalize(res);
   DROP2;
}
//*****************************************************************************
void N_Integer_bitRight(register CThread *thread)
//**********************************************
{
register MYINT right;
MyNatural *n, *res;
int l, cshift, cright;

   res = (MyNatural *) TOP;
   n = (MyNatural *) SECOND;
   right = TOCINT(THIRD);
   res->sign = n->sign;

   l = n->free - n->buf;
   cshift = right / NATURAL_SHIFT;
   cright = right % NATURAL_SHIFT;
   if (l < cshift) { THIRD = (Cell) OZERO; DROP2; return; }
   
   myNaturalRight(res->buf, n->buf + cshift, l - cshift, cright);
   THIRD = (Cell) normalize(res);
   DROP2;
}

//*****************************************************************************
void N_Integer_asFloat(CThread *thread)
//*************************************
{
register MYINT i, len;
MYFLOAT f;
MyFloat *res;
MyNatural *n;

   i = (MYINT) SECOND;
   res = (MyFloat *) TOP; 

   if (ISINT(i)) { res->f = (MYFLOAT) TOCINT(i); } 
   else {
      n = (MyNatural *) i;
      if (n->free - n->buf > 30) { res->f = HUGE_VALF * n->sign; }
      else {
         f = 0.0;
         len = n->free - n->buf;
         for(i = len - 1; i >= 0; i--) {
            f *= (MYFLOAT) NATURAL_BASE; 
            f += (MYFLOAT) n->buf[i];
            }
         if (n->sign == -1) f = -f;
         res->f = f;
         }
      }
   SECOND = TOP; DROP
}
//*****************************************************************************


#ifdef __FLOAT_ENABLE__

//*****************************************************************************
// Float METHODS
//*****************************************************************************

//*****************************************************************************
void N_Float_eq(register CThread *thread)
//*****************************************
{
MyFloat *f, *o;
MYFLOAT diff;

   f = (MyFloat *) TOP; o = (MyFloat *) SECOND; DROP;
   diff = o->f - f->f;
   if (diff < 0) diff = -diff;
   // dbg("f = %g, g = %g diff = %g, e = %g\n", f->f, o->f, diff, EPSILON);
   if (diff >= EPSILON ) { TOP = (Cell) OFALSE; } else { TOP = (Cell) OTRUE; } 
}
//*****************************************************************************
void N_Float_cmp(register CThread *thread)
//***************************************
{
MyFloat *f, *o;

   f = ((MyFloat *) TOP); DROP; 
   o = (MyFloat *) TOP;
   if (o->f <= f->f) { TOP = (Cell) OTRUE; } else { TOP = (Cell) OFALSE; }
}
//*****************************************************************************
void N_Float_rand(register CThread *thread)
//****************************************
{
   ((MyFloat *) TOP)->f = ((MYFLOAT) rand() + 1) / (((MYFLOAT) RAND_MAX) + 2);
}
//*****************************************************************************
void N_Float_asInteger(register CThread *thread)
//**********************************************
{
char *ds;
MYFLOAT f;
MYLONG *l;

   f = ((MyFloat *) TOP)->f; DROP
   if (f > __LONG_LONG_MAX__ || f < - __LONG_LONG_MAX__ ) 
      sysExceptionFromC(thread, "Float is too big to convert to integer");

   ds = ((char *) (DS - 2)) - sizeof(MYLONG); 
   l = (MYLONG *) ds; *l = (MYLONG) f;
}
//*****************************************************************************

#endif



//*****************************************************************************
// Support to dynamic libraries
//*****************************************************************************

#ifdef __DYN_ENABLE__

//*****************************************************************************
void N_OSLib_initialize(register CThread *thread)
//***********************************************
{
MyOSLib *oslib;
MyString *filename;
MYINT type;
void *handle;

   oslib = (MyOSLib *) TOP;
   type = (MYINT) SECOND;

   filename = (MyString *) THIRD;
   checkType(thread, filename, String);

   handle = (void *) LibLoad((char *) filename->buf); 
   if (handle == LibLoadError) sysExOS(thread);

   oslib->handle = handle;
   oslib->type = TOCINT(type);
   oslib->filename = filename;
   DROP3; 
}
//*****************************************************************************
void N_OSProc_initialize(register CThread *thread)
//************************************************
{
MyOSProc *osproc;
MyString *procname;
void *handle;
MyOSLib *lib;
vint nbParams;
Class *cl;

   osproc = (MyOSProc *) TOP;
   procname = (MyString *) SECOND;
   checkType(thread, procname, String);

   cl = (Class *) THIRD;
   nbParams = (MYINT) FOURTH;
   lib = (MyOSLib *) FIFTH;

   if (cl != NULL && cl != Integer && cl != String)
      sysExceptionFromC(thread, "Unsupported external return type");

   handle = (void *) ProcLoad(lib->handle, (char *) procname->buf); 
   if (handle == ProcLoadError) sysExOS(thread);

   osproc->handle = handle;
   osproc->lib = lib;
   osproc->procname = procname;
   osproc->nbParam = TOCINT(nbParams) * sizeof(Cell);
   osproc->clReturn = cl;
// dbg("Lib = %x, ret = %x <%s>\n", oslib, ((MyOSLib *) oslib)->handle, filename->buf);
   DROP5; 
}
//*****************************************************************************

#endif


#ifdef __DBG__

//*****************************************************************************
// Small disassembler
//*****************************************************************************

//*****************************************************************************
Cell N_st(register CThread *thread)
//*********************************
{
int i = 0;
Cell *ds, c;
MyObject *o;
unsigned char *pstate;

#define PERSISTENT(o)   (o->objptr & MSKPERSISTENT)

   ds = DS;
   printf("\n");
   if (ds > thread->dsbase) dbg("ds=%x, dsbase=%x:Stack overflow", 
      (vint) ds, (vint) thread->dsbase);

   while(ds < thread->dsbase) {
      o = (MyObject *) *ds;
      if (ISNIL(o) || ISINT(o)) { c = NULL; } 
      else { c = (Cell) o->objptr; }

      printf("%p %p [%2u] %8x %8x %s", thread->currentTask, ds, i, 
         (int) (sint) o, (int) (sint) c, symbuf(objectClass((MyObject *) o)->name));

      if (c != NULL && !(o->objptr & MSKPERSISTENT) && !(o->objptr & MSKNOTNEW)) { 
         pstate = (uchar *) *(((uchar **) o)-1); 
         printf(", state = %x", *pstate); 
         }
      if (isKindOf(o, String) == OK)  printf(" --> %s", ((MyString *) o)->buf);
      if (isKindOf(o, Integer) == OK) printf(" --> %10d", TOCINT(o));
      if (isKindOf(o, Word) == OK)    printf(" --> %s", symbuf(((MyWord *) o)->name));
      if (isKindOf(o, Symbol) == OK)  printf(" --> %s", symbuf(o));
      printf("\n");

      i++; ds++;
      }

   if (i == 0) printf("Empty\n");

   return NULL;
}
//*****************************************************************************
void deb_frame(CThread *thread, PCell first, PCell last)
//******************************************************
{
   while(first < last) {

#ifdef __x8632__
      dbg("Thread %x : Addr = %x, value = %x / %d\n", 
#endif

#ifdef __x8664__
      dbg("Thread %x : Addr = %I64d, value = %I64d / %I64x\n", 
#endif
         (vint) thread, (vint) first, (vint) *first, (vint) *first);

      first++;
      }
}
//*****************************************************************************
void deb_allframe(CThread *thread)
//********************************
{
   deb_frame(thread, thread->rsOforth, thread->rsbase);
}
//*****************************************************************************

char *Reg[16] = { 
    "(eax)", "(ecx)", "(edx)", "(ebx)", "(esp)", "(ebp)", "(esi)", "(edi)" ,
    "eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi" };

//*****************************************************************************
void dump(int addr, char *inst, ...)
//**********************************
{
va_list ap;

   va_start(ap, inst);
   printf("%8X : ", addr);
   vfprintf(stdout, inst, ap);
   printf("\n");
}
//*****************************************************************************
void decoderm(vint addr, char *op, uchar **pcode, vint dir)
//*********************************************************
{
vint rm, mode, regs, regt, regswap, size, val;
char scale;
   
   size = 2;
   rm   = (vint) *(*pcode + 1);
   mode = rm & 0xC0 ;                  // 0b11000000
   regs = ((0x38 & rm) >> 3) + 8;      // 0b00111000
   regt = 0x07 & rm;                   // 0b00000111

   // dbg("rm = %x, mode = %x regs = %d, regt = %d, dir = %d\n", rm, mode, regs, regt, dir);

   if (mode == 0x00 && regt == 0x05) {   // 32 bits displacement
      val = (vint) (* ((vint *) (*pcode + 2)));
      dump(addr, "%s [%x], %s", op, val, Reg[regs]); 
      size += sizeof(Cell);
      }

   else {
      if (mode == 0x40) {           // 0b01000000
         size++;
         if (regt != 4) scale = (*(* ((char **) pcode) + 2)) ;
         else { scale = (*(*((char **) pcode) + 3)); size++; }
         }
      else  { if (regt == 4) size++; }
      
      if (rm & 0x80) { regt += 8; }    //0b10000000
      if (dir == 1) { regswap = regs; regs = regt; regt = regswap; }

      if (mode == 0x40) {        //0b01000000
         if (regs < 8) dump(addr, "%s %d%s, %s", op, scale, Reg[regs], Reg[regt]);
         else          dump(addr, "%s %s, %d%s", op, Reg[regs], scale, Reg[regt]);
         }
      else { dump(addr, "%s %s, %s", op, Reg[regs], Reg[regt]); }
      }
   *pcode += size;
}
//*****************************************************************************
void disassemble(CThread *thread, uchar *code)
//********************************************
{
vint c, d, r;
uchar *base;
char v, w;
int add, add2, val;

   base = code;
   printf("\n");
   while(1) {
      r = code - base;
      c = *code;
      if (c == 0) return;
      d = *(code + 1);
      w = code[1];
      v = code[2];
      val = *((int *) (code + 1));
      add = *((int *) (code + 2));
      add2 = *((int *) (code + 3));

      printf("%2x:%2x:%2x:%2x:%2x: ", c, d, *(code + 2), *(code + 3), *(code + 4));

      if (c == 0x01) { decoderm(r, "add", &code, 0); continue; }
      if (c == 0x03) { decoderm(r, "add", &code, 1); continue; }

      if (c == 0x0F) {
         if (d == 0x84) { dump(r, "je %X", r + add + 6); code += 6; continue; }
         if (d == 0x85) { dump(r, "jne %X", r + add + 6); code += 6; continue; }
         if (d == 0x8E) { dump(r, "jle %X", r + add + 6); code += 6; continue; }
         }
      
      if (c == 0x29) { decoderm(r, "sub", &code, 0); continue; }
      if (c == 0x2B) { decoderm(r, "sub", &code, 1); continue; }
      if (c == 0x31) { decoderm(r, "xor", &code, 1); continue; }
      if (c == 0x39) { decoderm(r, "cmp", &code, 0); continue; }
      if (c == 0x3B) { decoderm(r, "cmp", &code, 1); continue; }

      if (c >= 0x48 && c <= 0x4F)
         { dump(r, "dec %s", Reg[c - 0x48 + 8]); code++; continue; }

      if (c >= 0x50 && c <= 0x57) 
         { dump(r, "push %s", Reg[c - 0x50 + 8]); code++; continue; }

      if (c >= 0x58 && c <= 0x5F) 
         { dump(r, "push %s", Reg[c - 0x57 + 8]); code++; continue; }

      if (c == 0x6A) { dump(r, "push %x", d); code += 2; continue; }

      if (c == 0x74) { dump(r, "je %X",  r + w + 2); code += 2; continue; }
      if (c == 0x75) { dump(r, "jne %x", r + w + 2); code += 2; continue; }
      if (c == 0x7E) { dump(r, "jle %X", r + w + 2); code += 2; continue; }

      if (c == 0x81) {
         if (d >= 0x20 && d <= 0x28)
            { dump(r, "and %x, %s", add, Reg[d - 0x20 + 8]); code += 6; continue; }

         if (d == 0x7E) { dump(r, "cmp %x, %d(esi)", add2, v); code += 7; continue; }

         if (d >= 0xC0 && d <= 0xC7)
            { dump(r, "add %x, %s", add, Reg[d - 0xC0 + 8]); code += 6; continue; }

         if (d >= 0xE0 && d <= 0xE7)
            { dump(r, "and %x, %s", add, Reg[d - 0xE0 + 8]); code += 6; continue; }

         if (d >= 0xE8 && d <= 0xEF)
            { dump(r, "sub %x, %s", add, Reg[d - 0xE8 + 8]); code += 6; continue; }

         if (d >= 0xF0 && d <= 0xF7) 
            { dump(r, "xor %x, %s", add, Reg[d - 0xF0 + 8]); code += 6; continue; }

         if (d >= 0xF8 && d <= 0xFF) 
            { dump(r, "cmp %x, %s", add, Reg[d - 0xF8 + 8]); code += 6; continue; }
         }

      if (c == 0x83) {
         if (d <= 0x07) { dump(r, "add %x, %x",      code[2], Reg[ d + 8]); code += 3; continue; }
         if (d == 0x3A) { dump(r, "cmp %x, (edx)",   code[2]); code += 3; continue; }
         if (d == 0x44) { dump(r, "add %x, %d(esp)", code[4], code[3]); code += 5; continue; }
         if (d == 0x6C) { dump(r, "sub %x, %d(esp)", code[4], code[3]); code += 5; continue; }
         if (d == 0x7E) { dump(r, "cmp %x, %d(esi)", code[3], code[2]); code += 4; continue; }

         if (d >= 0xC0 && d <= 0xCF)
            { dump(r, "add %x, %s", code[2], Reg[d - 0xC0 + 8]); code += 3; continue; }

         if (d >= 0xE8 && d <= 0xEF)
            { dump(r, "sub %x, %s", code[2], Reg[d - 0xE8 + 8]); code += 3; continue; }
      
         if (d >= 0xF8 && d <= 0xFF) 
            { dump(r, "cmp 0, %s", Reg[d - 0xF8 + 8]); code += 3; continue; }
         }

      if (c == 0x85) { decoderm(r, "test", &code, 0); continue; }
      if (c == 0x89) { decoderm(r, "mov", &code, 0); continue; }
      if (c == 0x8A) {
         if (d == 0x0B) { decoderm(r, "movb", &code, 0); continue; }
         if (d == 0x1C) { dump(r, "movb (edx, ecx), bl", v); code += 3; continue; }
         }

      if (c == 0x8B) { decoderm(r, "mov", &code, 1); continue; }

      if (c == 0x90) { return; }

      if (c >= 0xB8 && c <= 0xBF)
         { dump(r, "mov %x, %s", val, Reg[c - 0xB8 + 8 ]); code += 5; continue; }

      if (c == 0xC1) {
         if (d >= 0x20 && d <= 0x27) { dump(r, "shl %x, %s", code[2], Reg[d-0x20+8]); code += 3; continue; }
         if (d >= 0xF8 && d <= 0xFF) { dump(r, "sar %x, %s", code[2], Reg[d-0xF8+8]); code += 3; continue; }
         }

      if (c == 0xC3) { dump(r, "ret"); code++; continue; }

      if (c == 0xC6) {
         if (d == 0x44) { dump(r, "mov %x, %d(esp)", *(code + 4), code[3]); code += 5; continue; }
         }

      if (c == 0xC7) {
         if (d <= 0x07) { dump(r, "mov %x, %s", add, Reg[d]); code += 6; continue; }
         
         if (d == 0x40) { dump(r, "mov %x, %d(eax)", add2, v); code += 7; continue; }
         if (d == 0x44) { dump(r, "mov %x, %d(esp)", *((int *) (code + 4)), code[3]); code += 8; continue; }
         if (d == 0x46) { dump(r, "mov %x, %d(esi)", *((int *) (code + 3)), code[2]); code += 7; continue; }
         if (d == 0x47) { dump(r, "mov %x, %d(edi)", add2, v); code += 7; continue; }
         }

      if (c == 0xE8) { dump(r, "call %X", (int) r + 5 + val); code += 5; continue; }
      if (c == 0xE9) { dump(r, "jmp  %X", (int) r + 5 + val); code += 5; continue; }

      if (c == 0xEB) { dump(r, "jmp  %X", (int) r + 2 + d);   code += 2; continue; }

      if (c == 0xF7) { 
         if (d >= 0x00 && d <= 0x07) { dump(r, "test %s, %x", Reg[d], add); code += 6; continue; }
         if (d >= 0x40 && d <= 0x47) { dump(r, "test %d%s, %x", v, Reg[d-0x40], add); code += 6; continue; }
         if (d >= 0xC0 && d <= 0xC7) { dump(r, "test %s, %x", Reg[d-0xC0+8], add); code += 6; continue; }
         }

      if (c == 0xFF) {
         if (d == 0x37) { dump(r, "push (edi)"); code += 2; continue; }
         if (d == 0x77) { dump(r, "push %d(edi)", v); code += 3; continue; }
         if (d == 0xD2) { dump(r, "call *edx"); code += 2; continue; }
         if (d == 0xD3) { dump(r, "call *ebx"); code += 2; continue; }
         if (d == 0xE1) { dump(r, "jmp *ecx"); code += 2; continue; }
         }

      sysExParamFromC(thread, "Unknow X86 instruction", (void *) TOOINT(c));
      }
}
//*****************************************************************************
void N_Function_see(CThread *thread)
//**********************************
{
register Func *f;

   f = (Func *) TOP;
   // dbg("Dissasembling function %s at %x\n", symbuf(f->name), f->code);
   disassemble(thread, (uchar *) f->code);
   DROP;
}
//*****************************************************************************
void N_Message_see(CThread *thread)
//*********************************
{
MyMessage *m;
Class *cl;
MyMethod *n;

   m = (MyMessage *) TOP;
   cl = (Class *) SECOND;
   n = findMethod(m, cl);
   if (n == NULL) sysExceptionFromC(thread, "No implementation");
   disassemble(thread, (uchar *) n->code);
   DROP2;
}
//*****************************************************************************
void N_Message_seecl(CThread *thread)
//**********************************
{
MyMessage *m;
Class *cl;
MyMethod *n;

   m = (MyMessage *) TOP;
   cl = (Class *) SECOND;
   n = findClassMethod(m, cl);
   if (n == NULL) sysExceptionFromC(thread, "No implementation");
   disassemble(thread, (uchar *) n->code);
   DROP2;
}
//*****************************************************************************

#endif


