//*****************************************************************************
//*****************************************************************************
// OFORTH : Header file
//*****************************************************************************
//*****************************************************************************

/*
  Copyright (c) 2014, 2017 Franck Bensusan. All rights reserved.

  This file is part of Oforth sources engine.

  Oforth is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation, either version 3
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, see http://www.gnu.org/licenses/.
*/

#ifndef _OFORTH
#define _OFORTH

// Oforth Version
#define OFORTH_VERSION  "V1.2.0"

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>
#include <unistd.h>
#include <limits.h>
#include <locale.h>
#include <sys/time.h>
#include <sys/timeb.h>

typedef int *           Cell;
typedef int **          PCell;
typedef PCell           PTR;
typedef unsigned char   uchar;
typedef intptr_t        sint;
typedef uintptr_t       vint;


//*****************************************************************************
//*****************************************************************************
// Processor specific
//*****************************************************************************
//*****************************************************************************

//*****************************************************************************
// All 32bits
//*****************************************************************************
#if defined(__x8632__)
#define __OM32__

typedef int32_t         MYINT;
typedef int64_t         MYLONG;
typedef double          MYFLOAT;

#define MAX_CINT        0x3FFFFFFF
#define MIN_CINT        -0x3FFFFFF
#define CFLOAT_MAX      1073741823.0
#define CELLSHIFT       2
#define CELLBOUND       0x3
#define TWOC            (2*sizeof(Cell))

#define NATURAL_BASE    0x40000000
#define NATURAL_LOW     0x3fffffffL
#define NATURAL_SHIFT   30
#define NATURAL_PRINT   1000000000
#define NATURAL_PRINTS  9

#define MSKCLASSID      0x00000FFF
#define MSKSTATE        (~MSKCLASSID)
#define MSKATT          0x00001000
#define MSKSIZEALL      0x0003F000
#define MSKATTHIDDEN    0x00040000
#define MSKSIZEHIDDEN   0x003C0000
#define MSKNOTATTGC     0x00400000
#define MSKMODEINIT     0x00800000
#define MSKMODERUN      (~MSKMODEINIT)
#define MSKMUTABLE      0x01000000
#define MSKNOTMUTABLE   (~MSKMUTABLE)
#define MSKALLOC        0x02000000
#define MSKBUFFER       0x04000000
#define MSKPERSISTENT   0x08000000
#define MSKNPERSISTENT  (~MSKPERSISTENT)
#define MSKARRAYGC      0x10000000
#define MSKFINALIZE     0x20000000
#define MSKWORD         0x40000000
#define MSKNOTNEW       (MSKALLOC | MSKWORD)
#define MSKFREE         0x80000000

#define MSKSIZE         12
#define MSKHSIZE        18
#define MSKSIZESHIFT    12

#endif 

//*****************************************************************************

//*****************************************************************************
// Intel 32bits
//*****************************************************************************
#ifdef __x8632__

#define OPRET           0xC3 
#define PauseCpu        asm volatile("pause\n": : :"memory");
#define Barrier         asm volatile("": : : "memory");

#endif

//*****************************************************************************

//*****************************************************************************
// All 64 bits
//*****************************************************************************
#if defined(__x8664__)
#define __OM64__
typedef int64_t         MYINT;
typedef int64_t         MYLONG;
typedef double          MYFLOAT;

#define CELLSHIFT       3

#define QMEM            0x4000000000000000U 

#define CNATIVEMSK      0x0000000000000001U
#define CNATIVEADDR     0xFFFFFFFFFFFFFFFEU


#endif


//*****************************************************************************
// Intel 64 bits
//*****************************************************************************
#ifdef __x8664__

#define Barrier         asm volatile("": : : "memory");
#define PauseCpu        asm volatile("pause\n": : :"memory");

#endif

//*****************************************************************************

//*****************************************************************************
//*****************************************************************************
// OS Specific
//*****************************************************************************
//*****************************************************************************
#if defined(__OWIN32__) || defined(__OWIN64__)

#define FTOOINT(x)      ( ((vint) x) + 1)
#define OINTTOF(x)      ( ((vint) x) - 1)

// Windows
//*****************************************************************************
#define __OWIN__  
#undef  _WIN32_WINNT
#define _WIN32_WINNT     0x0501
#include <ws2tcpip.h>
#include <winsock2.h>
#include <windows.h>
#include <mmsystem.h>
#include <winbase.h>
#include <winuser.h>

#define PATH_FORMAT     ';'
#define LONG_FORMAT     "I64d"
#define LONG_SCAN       "%I64d"
#define myerrno          WSAGetLastError()

// Dynamic librairies
#ifdef __DYN_ENABLE__

#define LibLoad(x)       LoadLibrary(x)  
#define LibLoadError     NULL
#define ProcLoad(x, y)   GetProcAddress(x, y)
#define ProcLoadError    NULL

#endif

// Threads
#ifdef __MTHREAD_ENABLE__

#define ThreadHandle     int
#define THREAD_SPINLOCK   PauseCpu
#define THREAD_YIELD      Sleep(0)

#endif

#define THREAD_SLEEP      Sleep(1)     

// Memory
DWORD prevprotect;
#define AllocRaw(x,y)    VirtualAlloc(0, x, MEM_COMMIT, y)
#define FreeMem(x, y)    VirtualFree(x, 0, MEM_RELEASE)
#define PageProtect(x,y) VirtualProtect(x, OPAGESIZE, y, &prevprotect)
#define PageProtError    0

// Sockects
#ifdef __TCP_ENABLE__

WSADATA wsadata;
#define CWOULDBLOCK      WSAEWOULDBLOCK
#define CWOULDBLOCKA     WSAEWOULDBLOCK
#define COMPINTR         WSAEINTR
#define CONNECTERROR     
#define SockInit         WSAStartup(MAKEWORD(1, 1), &wsadata)
#define SockInitOk       0    
#define SockClean        WSACleanup();
#define SockCleanOk      0
#define SockClose(x)     closesocket(x)
#define SockError        SOCKET_ERROR
#define SockCloseAccept  WSAENOTSOCK
#define SockInterruptAccept WSAEINTR
#define SocketSetNonBlocking(s,x)  blockMode=1; x = ioctlsocket(s, FIONBIO, &blockMode)
#define SocketSetBlocking(s, x)    blockMode=0; x = ioctlsocket(s, FIONBIO, &blockMode)
#define SOL_NODELAY      SOL_SOCKET

#ifndef IPV6_V6ONLY
#define IPV6_V6ONLY 27
#endif

#endif      // __TCP_ENABLE__

// Console
extern int _kbhit(void);

#define CheckInputConsole _kbhit()
#define NoInputConsole   0
HANDLE STDIN, STDOUT;

#endif // Windows

//*****************************************************************************
// Linux & Mac
//*****************************************************************************
#if defined(__OLNX32__) || defined (__OLNX64__) || defined(__OMAC32__)
#define __OLNX__
#include <errno.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <pthread.h>
#include <termios.h>
#include <time.h>
#include <sys/mman.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <dlfcn.h>

#define PATH_FORMAT     ':'
#define LONG_FORMAT     "lld"
#define LONG_SCAN       "%lld"
#define myerrno          errno

struct termios original_ts;

// Dynamic librairies
#ifdef __DYN_ENABLE__
#define LibLoad(x)       dlopen(x, RTLD_LAZY)
#define LibLoadError     NULL
#define ProcLoad(x, y)   dlsym(x, y)
#define ProcLoadError    NULL
#endif

// Threads
#ifdef __MTHREAD_ENABLE__
pthread_attr_t commonAttr;
#define ThreadHandle     pthread_t *
#define ThreadSleepMicro  10000
#define THREAD_SPINLOCK   PauseCpu
#define THREAD_YIELD      sched_yield();
#endif

#define THREAD_SLEEP      sleepMicroSecond(ThreadSleepMicro);

// Memory
#define PAGE_READWRITE         PROT_READ | PROT_WRITE
#define PAGE_EXECUTE           PROT_EXEC | PROT_READ
#define PAGE_EXECUTE_READWRITE PROT_EXEC | PROT_READ | PROT_WRITE

#if defined(__OLNX32__) || defined (__OLNX64__)
#define AllocRaw(x,y)    mmap(NULL, x, y, MAP_ANONYMOUS | MAP_PRIVATE, -1, 0)
#endif

#if defined(__OMAC32__) || defined (__OMAC64__)
#define AllocRaw(x,y)    mmap(NULL, x, y, MAP_ANONYMOUS | MAP_PRIVATE, -1, 0)
#endif

#define FreeMem(x, y)    munmap(x, y)
#define PageProtect(x,y) mprotect(x, OPAGESIZE, y)
#define PageProtError    -1

// Sockects
#ifdef __TCP_ENABLE__
#define INVALID_SOCKET   -1
#define CWOULDBLOCKA     EAGAIN
#define CWOULDBLOCK      EAGAIN
#define COMPINTR         EINTR
#define CONNECTERROR     ECONNREFUSED
#define SockInit         0
#define SockInitOk       0    
#define SockClean        0
#define SockCleanOk      0
#define SockError        -1
#define SockCloseAccept  EBADF
#define SockInterruptAccept EINTR
#define SockClose(x)     close(x)
#define SocketSetNonBlocking(s, x)  blockMode = 1; x = ioctl(s, FIONBIO, &blockMode)
#define SocketSetBlocking(s, x)     blockMode = 0; x = ioctl(s, FIONBIO, &blockMode)
#define SOL_NODELAY      SOL_TCP
#endif   // __TCP_ENABLE


#endif // Linux

//*****************************************************************************
// OS Specific
//*****************************************************************************
#if defined(__OLNX32__)

// Files id
#define FTOOINT(x)      ( ((vint) x) + 1)
#define OINTTOF(x)      ( ((vint) x) - 1)

#endif // MAC

#if defined(__OMAC32__)

// Files id
#define FTOOINT(x)      ( ((vint) x) + 1)
#define OINTTOF(x)      ( ((vint) x) - 1)

#endif // MAC

//*****************************************************************************


//*****************************************************************************
//*****************************************************************************
// Commons
//*****************************************************************************
//*****************************************************************************

typedef MYLONG        TICK;

#define FIRST_PAGE    0
#define NEXT_RAW      1
#define NEXT_OBJECT   2

#define DICTIONARY    0
#define HEAP          1

#define LockCmpExchange(x, old, new) __sync_val_compare_and_swap(&(x), old, new)
#define LockAdd(x, y)                __sync_fetch_and_add(&(x), y)
#define LockSub(x, y)                __sync_fetch_and_sub(&(x), y)
#define LockInc(x)                   LockAdd(x, 1)
#define LockDec(x)                   LockSub(x, 1)

#ifdef __MTHREAD_ENABLE__

#define LockRes(r)      \
    while(1) {          \
       if (LockCmpExchange((r)->thread, NULL, thread) == NULL) break;    \
       while((r)->thread != NULL) { THREAD_SPINLOCK; }                   \
       }

#define RelRes(r)  Barrier; (r)->thread = NULL;

#else
#define LockRes(r) ;
#define RelRes(r) ;
#endif


#define AllocMem(x)           AllocRaw(x, PAGE_READWRITE)
#define AllocPage             AllocMem(OPAGESIZE)
#define AllocCode             AllocRaw(OPAGESIZE, PAGE_EXECUTE)

#define PageProtectExW(x)     PageProtect(x, PAGE_EXECUTE_READWRITE)
#define PageProtectEx(x)      PageProtect(x, PAGE_EXECUTE)

#define TICKINIT      20000
#define HEAP_WARN   1700000

//*****************************************************************************
#define CFRAMEUP        2U

#define OK              0
#define KO              1
#define DONE            2

#define nil             0x0
#define OZERO           1
#define ISNIL(x)        ((MYINT) (x) == nil)
#define ISINT(x)        ((MYINT) (x) & 1)
#define ISINTNIL(x)     (ISNIL(x) || ISINT(x))

#define OONE            (OZERO + 2)
#define OTWO            (OZERO + 4)
#define OTHREE          (OZERO + 6)
#define OFALSE          OZERO 
#define OTRUE           OONE 

#define EPSILON         1.0e-9

#define TOCINT(x)       ((MYINT) (((MYINT) (x) - 1) >> 1))
#define TOOINT(x)       ((MYINT) (((MYINT) (x) << 1) + 1))

//*****************************************************************************
// Oforth errors
#define ERESCLOSED       100
#define ETIMEOUT         101

#define ESOCKCREATE      300
#define ESOCKACCEPT      304
#define ESOCKOPT         305
#define ESOCKCONNECT     306
#define ESOCKHOST        307
#define ESOCKREAD        308
#define ESOCKWRITE       309

#define HASH_SIZE        149

#define DEFAULTDSMAIN    256
#define DEFAULTWORKERSDS 256
#define MAXPATH           32
#define MAX_LINE		 256
#define MAXATTRIBUTE      32
#define MAXNAME          256
#define MAXNAMEWORD      120
#define MAXPARAMS         32
#define MAXARGS           32

#define GCFREE          0x00
#define GCFINALIZE      0x01
#define GCNFINAL        0xFE
#define GCNOTUSED       0x02
#define GCNEW           0x04
#define GCUSED          0x08
#define GCUSEDIMMUT     0x10
#define GCPERSISTENT    0x20

#define DPPAGESIZE      (4096 - 2*sizeof(Cell))
#define OPAGESIZE       65536
#define KOBYPAGE        (OPAGESIZE / 1024)
#define MAXCODE         2048
#define WARNCODE        (MAXCODE - 20)
#define RAWSLOTS        (MAXPARAMS + 10)

#define MAX_MARKSTACK    512
#define MIN_CLASSES       64        // Number min of classes
#define MAX_CLASSES     1000        // Number max of classes

#define DS              thread->ds
#define FORMAT          thread->format
#define TOP             *(DS)
#define SECOND          *(DS+1)
#define THIRD           *(DS+2)
#define FOURTH          *(DS+3)
#define FIFTH           *(DS+4)
#define SIXTH           *(DS+5)

#define DROP            DS++;
#define DROP2           DS += 2;
#define DROP3           DS += 3;
#define DROP4           DS += 4;
#define DROP5           DS += 5;
#define DROP6           DS += 6;
#define PUSHDS(x)       *(DS-1) = (Cell) (x); DS--;
#define POPDS(x, t)     x = (t) *DS; DROP;
#define STACKSIZE       (thread->dsbase - thread->ds)

#define DATELOCAL       01
#define DATEGMT         00

#define SOCKREADABLE    0
#define SOCKWRITEABLE   1

#define RUNTIME_MODE    OZERO
#define COMPILE_MODE    OONE

#define CPREFIX         1
#define CSUFFIX         2

#ifdef __DBG__
   #define CHECKSTACK(x)   if (STACKSIZE < (x)) badStack(thread); 
#else
   #define CHECKSTACK(x) 
#endif


#ifdef __DBG__
   #define DEB             printf("Debug\n");
   #define dbg             printf
#endif


// *************************************************************************
// Pages
// *************************************************************************

typedef struct rph {
   vint nbFree;
   uchar *nextstate;
   struct rp *next;
   uchar *states;
   uchar *objects;
   vint  nbObj;
   } RawPageHeader;

#define RAWPAGEHEADERSIZE   sizeof(RawPageHeader)
#define RAWPAGEBODYSIZE     OPAGESIZE - RAWPAGEHEADERSIZE

typedef struct rp {
   RawPageHeader h;
   char body[RAWPAGEBODYSIZE];
   } RawPage;

typedef struct raw {
   struct raw *next;
   vint objptr;
   uchar raw;
   } Raw;

// *************************************************************************
// Basic objects
// *************************************************************************

typedef struct myobj {
   vint objptr;
   } MyObject;

typedef struct symb {
   vint objptr;
   struct symb *next;
   } MySymbol;

typedef struct blk {
   vint objptr;
   void (*code)();
   vint nbVars;
   vint varused;
   } MyBlock;

typedef struct blkv {
   vint objptr;
   void (*code)();
   vint nbVars;
   MyObject *vars;
   } MyClosure;

typedef struct flo {
   vint objptr;
   vint align;
   MYFLOAT f;
   } MyFloat;
#define FLOATSIZE    (sizeof(MyFloat) / sizeof(Cell))

typedef struct buffe {
   vint  objptr;
   uchar *buf;
   uchar *free1;
   uchar *alloc;
   } MyBuffer;

typedef struct stringe {
   vint  objptr;
   uchar *buf;
   uchar *free1;
   uchar *alloc;
   } MyString;

typedef struct array {
   vint objptr;
   Cell *buf;
   Cell *free;
   Cell *alloc;
   } MyArray;

typedef struct llo {
   vint objptr;
   vint *buf;
   vint *free;
   int  sign;
   } MyNatural;

typedef struct ex {
   vint objptr;
   MyString *message;
   void *object;
   } MyException;

typedef struct fileout {
   vint objptr;
   FILE *h;
   MyString *name;
   vint encoding;
   } MyFile;

typedef struct imppack {
   struct imppack *next;
   struct pck *package;
   } MyImportedPack;


#define IDPackage       8

typedef struct pck {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   MyArray *wrdlist;
   MyImportedPack *importedPackages;
   } MyPackage;
   
#define FILE_BIN     TOOINT(0)
#define FILE_TEXT    TOOINT(1)
#define FILE_UTF8    TOOINT(2)

#define FILEBEGIN    0
#define FILECURRENT  1
#define FILEEND      2

// Task states
#define TASK_RUN          0
#define TASK_YIELDING     1
#define TASK_PAUSING      3

#define IPSTORAGE         sizeof(struct sockaddr_storage)
#define FORMAT_SIZE       IPSTORAGE
#define FORMAT_MAX        (FORMAT_SIZE - 1)

typedef struct cpage {
   struct cpage *prev;
   uchar *target;
   uchar code[OPAGESIZE - sizeof(struct cpage *) - sizeof(uchar *)];
   } CodePage;

// Tasks
typedef struct task {
   Cell *rs;
   Cell *ds;
   vint state;
   Cell *ip;
   struct res *waitingResource;
   vint awakeOccured;
   MyNatural *natural;
   Cell *rsbase;
   struct task *resumableNext;
   struct task *waitPrev;
   struct task *waitNext;
   PCell vtvar;
   vint dsSize;
   MyObject *runnable;
   MyFloat f;
   vint rsMaxSize;
   Cell *dsbase;
   Cell *dsup;
   MySymbol **symbols;
   MyPackage *rootPackage;
   Cell *dp;
   vint dpCellLeft;
   Cell *dppages;
   CodePage *codePage;
   uchar *codeBuffer;
   TICK timeout;
   TICK sleep;
   uchar formatted[FORMAT_SIZE];
   } MyTask;

#define VM_RUN       0
#define VM_SNAPSHOT  1
#define VM_MARK      2
#define VM_SWEEP     3

// Finally, oforth vm
typedef struct oforthvm {
   vint state;
   struct cth *thread;
   vint nbVars;
   vint nbWorkers;
   vint nbWorkersMax;
   int initSnapshot;
   struct cth *currentThreadGC;
   vint nbWorkersToSynchronize;
   vint nbWorkersSnapshot;
   vint nbWorkersMarking;
   vint nbWorkersEndMarking;
   vint nbWorkersSweeping;
   TICK tick;
   } OforthVM;

// Thread state
#define THREAD_RUNNING      0x00000000
#define RESTRICT_GC         0x00100000
#define GLOBAL_GC           0x01000000
#define ASK_GC              0x00000001
#define ANSWERED_GC         0xFFFFFFFE

// Thread phases for GC
#define GC_NOGC             0x00
#define GC_SWEEPING         0x02
#define GC_SNAPSHOT         0x04
#define GC_MARKING          0x08
#define GC_MARKDONE         0x10

// Yielding parameters
#define  YIELD_TICKS        30000

// GC parameters
#define GC_TICK_MUTATOR     200
#define GC_TICK_GC        10000

typedef struct cth {
   Cell *ds;
   vint state;
   Cell *dsbase;
   RawPage **cur_obj;
   RawPage **all_obj;
   Raw **raws;
   Raw **obsolete;
   Cell *rsOforth;
   Cell *rsbase;
   Cell *rspausable;
   Cell *ippausable;
   MyTask *currentTask;
   int heapSize;
   Cell *dsupmax;
   OforthVM *vm;
   vint gcPhase;
   sint gcTicks;
   MyObject **markStack;
   MyObject **markStackBase;
   MyObject **markStackUp;
   vint *gcFreeIndex;
   vint gcSweepIndexObject;
   vint obsolete2;
   RawPage *gcSweepPage;
   RawPage *gcSweepPrevPage;
   Cell *obsolete3;
   Cell *obsolete4;
   vint obsolete5;
   char *persistentPage;
   vint persistentFree;
   struct cth *next;
   struct cth *thread;
   MyTask *resumableTasks;
   MyTask *resumeLast;
   MyTask *sleepingTasks;
   struct chan *channels;
   uchar *format;
   vint id;
   } CThread;

typedef Cell (*GetResource)(CThread *, struct res *, vint *);
typedef  int (*PutResource)(CThread *, struct res *, MyObject *, vint *);

typedef struct memlock {
   CThread *thread;
   } MemoryLock;

#define RECEIVE_MODE       0
#define SEND_MODE          1

#define ORECEIVE_MODE      OZERO
#define OSEND_MODE         OONE
// Resources
typedef struct res {
   vint objptr;
   CThread *thread;
   MyTask *waitNotEmpty;
   MyTask *waitNotFull;
   vint selectMode;
   int oclosed;
   } MyResource;

// Channels
typedef struct chan {
   vint objptr;
   CThread *thread;
   MyTask *waitNotEmpty;
   MyTask *waitNotFull;
   vint selectMode;
   int oclosed;
   Cell *buf;
   vint first;
   vint last;
   struct chan *next;
   struct chan *prev;
   vint size;
   vint maxSize;
   } MyChannel;
#define CHANNELSIZE   (sizeof(MyChannel) / sizeof(Cell))

// TCPSocket
typedef struct tsock {
   vint objptr;
   CThread *thread;
   MyTask *waitNotEmpty;
   MyTask *waitNotFull;
   vint selectMode;
   int oclosed;
   vint sock;
   vint sockv6;
   struct addrinfo *results;
   struct addrinfo *currentResult;
   uchar *remainingBuffer;
   vint remainingSize;
   MyString *port;
   } MyTCPSocket;

typedef struct tsocks {
   vint objptr;
   CThread *thread;
   MyTask *waitNotEmpty;
   MyTask *waitNotFull;
   vint selectMode;
   int oclosed;
   vint sock;
   vint sockv6;
   struct addrinfo *results;
   struct addrinfo *currentResult;
   uchar *remainingBuffer;
   vint remainingSize;
   MyString *port;
   vint mode;
   vint backlog;
   } MyTCPSocketServer;

typedef struct tsockc {
   vint objptr;
   CThread *thread;
   MyTask *waitNotEmpty;
   MyTask *waitNotFull;
   vint selectMode;
   int oclosed;
   vint sock;
   vint sockv6;
   struct addrinfo *results;
   struct addrinfo *currentResult;
   uchar *remainingBuffer;
   vint remainingSize;
   MyString *port;
   MyString *host;
   MyString *remoteport;
   } MyTCPSocketClient;

// Console
typedef struct cons {
   vint objptr;
   CThread *thread;
   MyTask *waitNotEmpty;
   MyTask *waitNotFull;
   vint selectMode;
   int oclosed;
   MyString *cmd;
   vint xcursor;
   void *history;
   } MyConsole;

typedef void (*fptr)(CThread *);
typedef void (fc)(register CThread *);
typedef void (ffinal)(CThread *, MyObject *);


// *************************************************************************
// Words
// *************************************************************************

typedef struct syma {
   vint objptr;
   struct symb *next;
   char *buf;
   } MySymAddr;
#define symbuf(s)    ((char *) (&(((MySymAddr *) (s))->buf)))

typedef struct wrd {
   vint objptr;
   struct wrd *next;
   MySymbol *name;
   } MyWord;

typedef struct tnote {
   vint objptr;
   sint level;
   void *whenAddr;
   } TryNode;

typedef struct func {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   void (*code)();               // Code pointer
   } Func;

typedef struct dual {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   void (*code)();               // Code pointer
   void (*compileCode)();        // Compile code pointer
   } FDual;

typedef struct tnodemethod {
   vint objptr;
   struct tnodemethod *next;     // Next implementation
   void (*code)();               // Code pointer
   struct tclass *implementor;   // Class or property implementation
   } MyMethod;
   
typedef struct tmessage {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   PCell vt;
   MyMethod *objectList;      
   MyMethod *classeList;    
   } MyMessage;

typedef struct tna {
   struct tna *next;
   MySymbol *name;
   struct tprop *property;
   vint isMutable;
   } Attribute;

typedef struct timpl {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   Attribute *attList;     // List of attributes of the property
   } Implementor;

typedef struct tprop {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   Attribute *attList;     // List of attributes of the property
   } MyProperty;

typedef struct pna {
   struct pna *next;
   MyProperty *prop;
   } MyNodeProp;

typedef struct tclass {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   Attribute *attList;              // List of attributes of the class
   fc *initFunction;
   ffinal *finalFunction;
   vint classPtr;
   vint size;                       // Size of instances (1+nbatt) of this class
   MyNodeProp *propList;
   struct tclass *super;            // Super class
   } Class;

typedef struct constant {
    vint objptr;
    struct MyWord *next;
    MySymbol *name;
    Cell value;
    } MyConstant;

typedef struct var {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   vint index;
   } MyVariable;

typedef struct oslib {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   void *handle;
   vint type;
   MyString *filename;
   } MyOSLib;

typedef struct osproc {
   vint objptr;
   struct MyWord *next;
   MySymbol *name;
   void *handle;
   vint nbParam;
   MyOSLib *lib;
   MyString *procname;
   Class *clReturn;
   } MyOSProc;

#endif // OFORTH

