//*****************************************************************************
//*****************************************************************************
// OFORTH : TASKS AND PARALLEL
//*****************************************************************************
//*****************************************************************************

/*
  Copyright (c) 2014, 2018 Franck Bensusan. All rights reserved.

  This file is part of Oforth sources engine.

  Oforth is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation, either version 3
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, see http://www.gnu.org/licenses/.
*/

#include "oforth.h"

#ifdef __OLNX__
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#define  SOCKADDR	struct sockaddr

extern pthread_attr_t commonAttr;

#endif

extern void sysExceptionFromC(CThread *thread, char *s);
extern void sysExParamFromC(CThread *thread, char *s, void *x);
extern void sysExOS(register CThread *thread);

extern void strcatc(char *str, char c);
extern RawPage *allocRawMem(register CThread *thread, register vint size);
extern PTR allocMem(CThread *, vint, char **);
extern void freeMem(CThread *, PTR);
extern PCell allocDictCell(CThread *thread, vint n);
extern void outOfMemory(CThread *thread);
extern void badParams(CThread *);
extern void badStack(CThread *thread);
extern Class *objectClass(MyObject *o);
extern vint isKindOf(MyObject *o, Class *parent);
extern void checkType(CThread *thread, void *o, Class *cl);
extern void checkTypeInteger(CThread *thread, vint i);
extern void bufAddChars(CThread *, MyBuffer *str, char *toadd, vint sizetoadd);

extern void GCHandleRequest(CThread *thread);
extern void sysThreadExecute(CThread *thread);
extern void sysThreadPauseTask(CThread *thread);
extern void checkImmutable(register CThread *thread, MyObject *o);

extern void deb_frame(CThread *thread, PCell first, PCell last);

// Threads & Tasks
extern OforthVM *ovm;
extern vint workerId;
extern MyTask *mainTask;
extern CThread *mainThread, *lastThread;
extern vint codeMem, dictMem;

extern Class *Object, *Word, *Integer, *Float;
extern Class *Buffer, *String;
extern Class *Task, *Block, *Console, *Channel;
extern Class *TCPSocket, *TCPSocketServer, *TCPSocketClient;

extern void setvar(CThread *thread, MyVariable *v, Cell value);

// Resources
extern vint genericVT;
extern MyTask *mainTask;
extern char *inputBuffer;
extern vint heapSizeKo;
extern vint mainDsSize;
extern int consoleMode;
extern MyConsole *mainConsole;
extern MyFile *stdoutFile;
extern Cell sysFileOutAdd(register CThread *thread);

Cell sysChannelGet(CThread *thread, MyResource *rc);

extern void GC(int type, TICK tick);

extern MyVariable *SOURCE, *SOURCELINE, *SOURCEINDEX;

extern vint gcTimer, gcMemory, gcTicksByStep, gcAppByStep, gcVerboseLevel;
extern TICK nextLocalGC, nextGlobalGC, nextInitTick;
extern vint lastNumberRuns;
extern vint GCLOCAL, GCGLOBAL;
extern vint maxMemory;

// SOCKETS
#define MODE_IPV4     1
#define MODE_IPV6     2
#define MODE_IPALL    3

void flushCurrentCmd(register CThread *thread);

//*****************************************************************************
// Times, ticks ans sleep
//*****************************************************************************

extern MYFLOAT tickFrequency;

#ifdef __OLNX__
//*****************************************************************************
void sleepMicroSecond(int x)
//**************************
// Sleep for x micro seconds
{
struct timespec req;  

   req.tv_sec  = x / 1000000;  
   req.tv_nsec = (x % 1000000) * 1000;

#if defined(__OLNX32__) || defined(__OLNX64)
   while(clock_nanosleep(CLOCK_MONOTONIC, 0, &req, &req) == -1);
#endif

#if defined(__OMAC32__) || defined(__OMAC64)
   while(nanosleep(&req, &req) == -1); 
#endif

}
#endif

//*****************************************************************************
MYLONG getTick(int mode, struct timeb *t)
//***************************************
{
MYLONG ul;
struct timeb tp;

   if (mode == KO || tickFrequency == 0.0) {
      ftime(&tp);
      ul = ((MYLONG) tp.time) * 1000000 + ((MYLONG) tp.millitm) * 1000;
      if (t != NULL) {
         t->dstflag = tp.dstflag;
         t->timezone = tp.timezone;
         }
      }
   else {
#ifdef __OWIN__
LARGE_INTEGER tick;
      QueryPerformanceCounter(&tick);
      ul = (MYLONG) ((double) tick.QuadPart * tickFrequency);
#endif

#if defined(__OLNX32__) || defined(__OLNX64__)
struct timespec current;
      clock_gettime(CLOCK_REALTIME, &current);
      ul = ((MYLONG) current.tv_sec) * 1000000 + ((MYLONG) current.tv_nsec) / 1000;
#endif

#if defined(__OMAC32__) || defined(__OMAC64__)
struct timeval current; 
      gettimeofday(&current, NULL);
      ul = ((MYLONG) current.tv_sec) * 1000000 + (MYLONG) current.tv_usec;
#endif
      }

   return ul;
}
//*****************************************************************************
void N_SystemLocalTime(register CThread *thread)
//**********************************************
{
struct timeb t;
char *ds;
MYLONG *l;

   PUSHDS(NULL); PUSHDS(NULL); 
   ds = ((char *) (DS - 2)) - sizeof(MYLONG); 
   l = (MYLONG *) ds; *l = getTick(KO, &t);

   if (t.dstflag == 0) { SECOND = (Cell) OFALSE; } else { SECOND = (Cell) OTRUE; }
   TOP = (Cell) TOOINT( - t.timezone);
}
//*****************************************************************************
void N_SystemTick(register CThread *thread)
//*****************************************
{
char *ds;
MYLONG *l;

   ds = ((char *) (DS - 2)) - sizeof(MYLONG); 
   l = (MYLONG *) ds; *l = getTick(OK, NULL);
}
//*****************************************************************************
void retrieveTimeout(CThread *thread, MyTask *t, MYINT ot)
//********************************************************
{
   if (ISNIL(ot)) { t->timeout = LLONG_MAX; }
   else {
      checkTypeInteger(thread, ot);
      t->timeout = TOCINT(ot) + 1;
      if (ot != OZERO) t->timeout += getTick(OK, NULL);
      }
}
//*****************************************************************************


//*****************************************************************************
// Tasks
//*****************************************************************************

//*****************************************************************************
int totalHeap()
//*************
{
CThread *th;
int heapSize = 0;

   for(th = mainThread; th != NULL; th = th->next) heapSize += th->heapSize;
   return heapSize;
}
//*****************************************************************************

#ifdef __DBG__

//*****************************************************************************
void printTask(MyTask *t, vint id, char *state)
//*********************************************
{
MyObject *m;

   m = t->runnable; 
   printf("%6d %9s ", id, state);
   if (isKindOf(m, Word)  == OK) printf("#%-39s\n", symbuf(((MyWord *) m)->name));
   if (isKindOf(m, Block) == OK) printf("%-40s\n", "aBlock");
}
//*****************************************************************************
void N_t(CThread *thread)
//***********************
{
CThread *th;
MyChannel *ch;
MyTask *t;
vint id;

   printf("Number of workers : %d\n", ovm->nbWorkers);
   printf("WORKER     STATE TASK\n");

   for(th = mainThread; th != NULL; th = th->next) {
      id = th->id;
      t = th->currentTask;
      if (t != NULL) printTask(t, id, "RUN");

      for(t = th->resumableTasks; t != NULL; t = t->resumableNext)
         printTask(t, id, "RESUMABLE");

      for(t = th->sleepingTasks; t != NULL; t = t->waitNext)
         printTask(t, id, "SLEEP");

      for(ch = th->channels; ch != NULL; ch = ch->next) {
         for(t = ch->waitNotEmpty; t != NULL; t = t->waitNext) 
             printTask(t, id, "WAIT");
         for(t = ch->waitNotFull ; t != NULL; t = t->waitNext) 
             printTask(t, id, "WAIT");
        }

      }
}
//*****************************************************************************

#endif


//*****************************************************************************
// Threads & tasks
//*****************************************************************************

MyTask *resourcePopTaskWaitingPut(register MyResource *);

//*****************************************************************************
Cell I_bye(CThread *thread)
//*************************
{

   if (thread->currentTask != NULL && thread->currentTask != mainTask) {
      sysExceptionFromC(thread, "Only main task can exit oforth");
      }

#ifdef __TCP_ENABLE__
int sockret = 0;

   sockret = SockClean;
   if (sockret != SockCleanOk)
      { printf("Sock clean failed...%d\n", myerrno); exit(KO); }
#endif

#ifdef __OLNX__
    if (consoleMode == OK) tcsetattr(fileno(stdin), TCSANOW, &original_ts);
#endif

   exit(TOCINT(SECOND));
}
//******************************************************************************

#ifdef __MTHREAD_ENABLE__

void sysThread(register CThread *thread);

//******************************************************************************
void newWorker(CThread *thread)
//*****************************
{
OforthVM *vm = thread->vm;
CThread *newThread;
ThreadHandle thh;
    
   newThread = (CThread *) allocRawMem(thread, sizeof(CThread));
   newThread->id = workerId; workerId++;
   // dbg("New worker %x, id = %d\n", newThread, newThread->id);
   lastThread->next = newThread;
   lastThread = newThread;
   newThread->next = NULL;
   vm->nbWorkers++;

   LockAdd(vm->nbWorkersToSynchronize, 1);

   newThread->vm      = vm;
   newThread->thread  = NULL;
   newThread->state   = THREAD_RUNNING;
   newThread->gcTicks = YIELD_TICKS;
   newThread->gcPhase = GC_NOGC;
   newThread->rsbase  = newThread->rsOforth = NULL; 
   newThread->persistentPage = NULL;
   newThread->persistentFree = 0;
   newThread->raws    = (Raw **)     allocDictCell(thread, RAWSLOTS); 
   newThread->all_obj = (RawPage **) allocDictCell(thread, MAXATTRIBUTE + 1); 
   newThread->cur_obj = (RawPage **) allocDictCell(thread, MAXATTRIBUTE + 1); 
   newThread->gcFreeIndex = (vint *) allocDictCell(thread, MAXATTRIBUTE + 1);
   newThread->heapSize = 0; 
   newThread->currentTask = NULL;
   newThread->resumableTasks = newThread->resumeLast = NULL; 
   newThread->sleepingTasks = NULL;
   newThread->channels = NULL;

   newThread->markStackBase = 
      (MyObject **) allocRawMem(thread, MAX_MARKSTACK * sizeof(Cell));
   newThread->markStack   = newThread->markStackBase;
   newThread->markStackUp = newThread->markStack + MAX_MARKSTACK - 1;

#ifdef __OWIN__
   thh = (int) _beginthreadex(NULL, OPAGESIZE, (void *) sysThread, newThread, 0, NULL);
#endif

#ifdef __OLNX__
   thh = malloc(sizeof(pthread_t));
   pthread_attr_setstacksize(&commonAttr, OPAGESIZE);
   pthread_create(thh, &commonAttr, (void * (*)(void *)) sysThread, newThread);
#endif

   if (thh <= 0) 
      sysExceptionFromC(thread, "System is unable to create a new worker");

   // dbg("New thread created : %x\n", newThread);
}

#endif

//*****************************************************************************
void awakeTask(register CThread *thread, register MyTask *t)
//**********************************************************
{
    t->resumableNext = NULL;

    LockRes(thread);
        if (thread->resumableTasks == NULL) { thread->resumableTasks = t; }
        else { thread->resumeLast->resumableNext = t; }
        thread->resumeLast = t;
    RelRes(thread);
}
//*****************************************************************************
void N_sched(CThread *thread)
//***************************
{
register MyTask *t;
MYINT dsSize;
MyNatural *n;
MyObject *runnable;
char *buf;

   CHECKSTACK(2)

   dsSize = (MYINT) TOP;
   checkTypeInteger(thread, dsSize);
   dsSize = (MYINT) TOCINT(dsSize);
   if (dsSize <= 10) dsSize = 10;

   runnable = (MyObject *) SECOND;
   if (ISINTNIL(runnable)) sysExceptionFromC(thread, "Bad value for runnable");

   if (runnable->objptr & MSKMUTABLE)
       sysExceptionFromC(thread, "Immutable rule violation");

   buf = (char *) allocMem(thread, 
       sizeof(MyTask) + 12 * sizeof(Cell) + dsSize * sizeof(Cell), NULL);

   t = (MyTask *) buf;
   t->state = TASK_RUN;
   t->ip    = NULL;
   t->waitingResource = NULL;
   t->waitPrev = t->waitNext = NULL;
   t->runnable = runnable;
   t->awakeOccured = KO;
   t->dsSize = TOCINT(dsSize);
   t->vtvar = (PCell) &genericVT;
   t->rs = t->rsbase = NULL;

   t->symbols = NULL;
   t->rootPackage = NULL;
   t->dp = t->dppages = NULL;
   t->dpCellLeft = 0;

   buf += sizeof(MyTask); 
   t->natural = n = (MyNatural *) buf ;
   n->objptr = (Integer->classPtr | MSKPERSISTENT) & MSKMODERUN;
   n->buf = (vint *) (buf + 4 * sizeof(Cell));
   n->free = n->buf + 1;

   buf += 8 * sizeof(Cell);
   t->dsup = (PCell) buf;
   t->dsbase = t->dsup + dsSize;
   *(t->dsbase) = *(t->dsbase+1) = NULL;
   t->ds = t->dsbase;

   t->rsMaxSize = 0;
   t->codePage = NULL;
   t->codeBuffer = NULL;

#ifdef __FLOAT_ENABLE__
   t->f.objptr = (Float->classPtr | MSKPERSISTENT) & MSKMODERUN;
#endif

   thread->gcTicks -= 1000;
   awakeTask(thread, t);
   DROP2;

   // printf("Schedule task %d %d, natural = %d %d, dsup = %d, ds = %d all = %d\n",
        // t, ((vint) t) + sizeof(MyTask), t->natural, t->natural->buf, t->dsup, t->ds,
       // ((vint) t) + sizeof(MyTask) + 12 * sizeof(Cell) + dsSize*sizeof(Cell));
}
//*****************************************************************************
void addSleepingTask(CThread *thread, MyTask *t)
//**********************************************
{
MyTask *s, *sprev;
MYLONG sleep;

   sleep = t->sleep;
   // dbg("add sleeping task %x, sleep = %I64d \n", t, sleep); 

   s = sprev = thread->sleepingTasks;
   while(s != NULL && s->sleep < sleep) { sprev = s; s = s->waitNext; }

   t->waitNext = s;
   if (s == thread->sleepingTasks) { thread->sleepingTasks = t; }
   else { sprev->waitNext = t; }

   // dbg("sleeping task added %x\n", t); 
}
//*****************************************************************************
void pauseTask(CThread *thread)
//*****************************
{
MyTask *t;
vint stackSize;
char *max;

    t = thread->currentTask; 
    t->ip = thread->ippausable;
    t->ds = DS;

    stackSize = (vint) thread->rsbase - (vint) thread->rspausable;

    if (stackSize > t->rsMaxSize) {
        if (t->rsMaxSize != 0) freeMem(thread, t->rs);
        t->rs = allocMem(thread, stackSize, &max);
        t->rsMaxSize = (vint) max - (vint) t->rs;
        }

    t->rsbase = t->rs + stackSize / sizeof(Cell);
    memcpy(t->rs, thread->rspausable, stackSize); 

    // dbg("Pausing task %p, it = %p, rsbase = %p\n", t, it, it->rsbase);
    thread->rspausable = thread->ippausable = NULL;
    sysThreadPauseTask(thread);
}
//*****************************************************************************
void resumeTask(register CThread *thread, register MyTask *t)
//***********************************************************
{
CodePage *cp, *cprem;
PCell dp, dpprev;

   // dbg("Thread %x %x : Resuming %x it = %x\n", thread->id, thread->state, t, it);

   thread->currentTask = t; 
   thread->dsbase  = t->dsbase;
   thread->dsupmax = t->dsup;
   thread->format  = t->formatted;

   while(t->state != TASK_RUN) { PauseCpu; }
   thread->ds      = t->ds;

   sysThreadExecute(thread); 
   thread->currentTask = NULL;

   // dbg("Thread %x : returning %x it = %x tit = %x rsbase = %x\n", thread->id, t, it, t->it, t->it->rsbase);

   
   if (t->state == TASK_PAUSING) { t->state = TASK_RUN; return; }

   if (t->state == TASK_YIELDING) {
      t->awakeOccured = OK;
      awakeTask(thread, t); 
      t->state = TASK_RUN;
      return; 
      }

   if (t == mainTask) I_bye(thread);

   // Free task : Remove code pages for this task
   cp = t->codePage;
   while(cp != NULL) {
      cprem = cp; cp = cp->prev;
      FreeMem(cprem, OPAGESIZE);
      LockSub(codeMem, OPAGESIZE); 
      }

   // Free task : remove dictioanry pages
   dp = t->dppages;
   while ( dp != NULL) {
      dpprev = dp; dp = (PCell) *dp; 
      // dbg("dpprev to free = %x, dp = %x\n", dpprev, dp);
      freeMem(thread, dpprev);
      }

   // Free task : Remove variables table 
   if (t->vtvar != (PCell) &genericVT) freeMem(thread, t->vtvar); 

   // Free task : Remove return stack, inner task and task itself
   if (t->rs != NULL) freeMem(thread, t->rs);
   freeMem(thread, (PCell) t);

   // dbg("Thread %x : task %x has stopped, state = %d\n", thread, t, t->state);
}
//*****************************************************************************
void checkEvents(CThread *thread, TICK now)
//*****************************************
{
MyTask *t;
int heap;

    // Check for sleeping tasks
    t = thread->sleepingTasks;
    while(t != NULL && t->sleep <= now ) {
        t->awakeOccured = OK; 
        t->sleep = now;
        awakeTask(thread, t); 
        t = t->waitNext;
        }
    thread->sleepingTasks = t;

    // Now, if main thread, check for garbage collector
    if (thread == mainThread && ovm->state == VM_RUN) {
        heap = totalHeap(); 
        if (heap < gcMemory) return;
        if (now > nextGlobalGC || heap > maxMemory) {
            // dbg("now = %I64d next = %I64d heap = %d max = %d\n", now, nextGlobalGC, heap, maxMemory);
            GC(GLOBAL_GC, now);
            nextLocalGC  = now + GCLOCAL;
            nextGlobalGC = now + GCGLOBAL;
            return;
            }

        if (now > nextLocalGC) {
            GC(RESTRICT_GC, now);
            nextLocalGC  = now + GCLOCAL;
            }
        }
}
//*****************************************************************************
void sysThread(register CThread *thread)
//**************************************
{
register MyTask *t;
OforthVM *vm = thread->vm;

#ifdef __MTHREAD_ENABLE__
CThread *th;
#endif

   while(1) {
      t = NULL;

      // If no more ticks while running, check events
      if (thread->state == THREAD_RUNNING && thread->gcTicks <= 0) {
         thread->gcTicks = YIELD_TICKS;
         checkEvents(thread, getTick(OK, NULL));
         }

      // If GC is running, handle a new GC step
      if (thread->state != THREAD_RUNNING) 
         { thread->gcTicks--; GCHandleRequest(thread); }

      // Now find a new task to resume

      if (thread->resumableTasks != NULL) {
          LockRes(thread);
              t = thread->resumableTasks;
              if (t != NULL) thread->resumableTasks = t->resumableNext;
          RelRes(thread);

          if (t != NULL) { resumeTask(thread, t); continue; }
          }

      // If no task to resume for this thread, retrieve a waiting task from another
#ifdef __MTHREAD_ENABLE__
      for( th = mainThread; th != NULL; th = th->next) {
         if (th == thread || th->resumableTasks == NULL) continue;

         LockRes(th);
            t = th->resumableTasks;
            if (t != NULL && t != mainTask ) th->resumableTasks = t->resumableNext;
            else { t = NULL; }
         RelRes(th);
         
         if (t != NULL) break;
         }
#endif
            
      if (t != NULL) { resumeTask(thread, t); continue; }

      // If GC is still running, loop
      thread->gcTicks = 0;
      if (thread->state != THREAD_RUNNING) continue;

      // dbg("Worker %d is sleeping\n", thread->id);
      LockSub(vm->nbWorkersToSynchronize, 1);

      THREAD_SLEEP;
      while (vm->state != VM_RUN && vm->state != VM_SWEEP) { THREAD_SLEEP; }

      // dbg("Worker %d is awake\n", thread->id);
      LockAdd(vm->nbWorkersToSynchronize,  1);
      }
}
//******************************************************************************
void N_yield(CThread *thread)
//***************************
{
MyTask *current;

   current = thread->currentTask;

   if (current->awakeOccured == OK) { current->awakeOccured = KO; }
   else { current->state = TASK_YIELDING; pauseTask(thread); }
}
//******************************************************************************
void N_sleep(CThread *thread)
//***************************
{
int ms;
TICK milli;
MyTask *current;

   current = thread->currentTask;

   if (current->awakeOccured == OK) { current->awakeOccured = KO; }
   else {
      ms = (int) TOP;
      checkTypeInteger(thread, ms);

      milli = ((TICK) TOCINT(ms)) * 1000U;
      current->sleep = milli + getTick(OK, NULL);
      addSleepingTask(thread, current);
      current->state = TASK_PAUSING;
      pauseTask(thread); 
      }

   DROP;
}
//*****************************************************************************


//*****************************************************************************
// Ressources
//*****************************************************************************

//*****************************************************************************
void sysResourceInit(register CThread *thread)
//********************************************
{
register MyResource *rc;

   rc = (MyResource *) TOP;
   rc->thread = NULL;
   rc->waitNotEmpty = rc->waitNotFull = NULL;
   rc->selectMode = KO;
   rc->oclosed = OFALSE;
   DROP;
}
//*****************************************************************************
void resourcePushWaitingTask(register MyTask **list, register MyTask *t)
//**********************************************************************
{
   t->waitPrev = NULL;
   t->waitNext = *list;
   if (t->waitNext != NULL) t->waitNext->waitPrev = t; 
   *list = t;
}
//*****************************************************************************
MyTask *resourcePopTaskWaitingPut(register MyResource *rc)
//********************************************************
{
register MyTask *t;

   t = (MyTask *) rc->waitNotFull; 

   if (t != NULL) { 
      if ((rc->waitNotFull = t->waitNext) != NULL) rc->waitNotFull->waitPrev = NULL;
      }

   return t;
}
//*****************************************************************************
MyTask *resourcePopTaskWaitingGet(register MyResource *rc)
//********************************************************
{
register MyTask *t;

   t = (MyTask *) rc->waitNotEmpty; 

   if (t != NULL) { 
      if ((rc->waitNotEmpty = t->waitNext) != NULL) rc->waitNotEmpty->waitPrev = NULL;
      }

   return t;
}
//************************************************************************************
Cell resourceGet(register CThread *thread, MyResource *rc,GetResource get,vint *perror)
//***********************************************************************************
{
register MyTask *current;
MyTask *t = NULL;
Cell e = NULL;

   current = thread->currentTask;

   if (current->awakeOccured == OK) { 
      current->awakeOccured = KO; 
      if (current->timeout <= current->sleep)
          { *perror = ETIMEOUT; goto LReturnGetEnd; }
      }

   LockRes(rc);
   e = get(thread, rc, perror);

   if (e != NULL) {
      t = resourcePopTaskWaitingPut(rc); 
      if (t != NULL) awakeTask(thread, t);
      goto LReturnGetRelease;
      }

   if ( *perror != 0) goto LReturnGetRelease;
   if (rc->oclosed == OTRUE)  { *perror = ERESCLOSED; goto LReturnGetRelease; }
   if (current->timeout == 1) { *perror = ETIMEOUT;   goto LReturnGetRelease; }

   current->waitingResource = rc;
   current->state = TASK_PAUSING;

   if (current->timeout == LLONG_MAX && rc->selectMode == KO) {
      resourcePushWaitingTask(&(rc->waitNotEmpty), current); 
      RelRes(rc);
      }
   else {
      RelRes(rc);
      current->sleep = getTick(OK, NULL) + 2000;   // Next check into 2 ms
      addSleepingTask(thread, current);
      }
   pauseTask(thread); 

LReturnGetRelease:
   RelRes(rc);

LReturnGetEnd:
   current->waitingResource = NULL;
   return e; 
}
//*****************************************************************************
int resourcePut(register CThread *thread, MyResource *rc, MyObject *o, PutResource put, vint *perror)
//*******************************************************************************
{
register MyTask *current;
MyTask *t = NULL;
int ret = KO;

   current = thread->currentTask;

   if (current->awakeOccured == OK) { 
      current->awakeOccured = KO; 
      if (current->timeout <= current->sleep) { *perror = ETIMEOUT; goto LReturnPut; }
      }

   if (rc->oclosed == OTRUE) { *perror = ERESCLOSED; goto LReturnPut; }

   LockRes(rc)

   ret = put(thread, rc, o, perror);

   if (ret == OK) { 
      t = resourcePopTaskWaitingGet(rc);
      if (t != NULL) awakeTask(thread, t);
      goto LReturnPutRelease;
      }

   if ( *perror != 0 ) goto LReturnPutRelease;
   if (current->timeout == 1) { *perror = ETIMEOUT; goto LReturnPutRelease; }

   current->waitingResource = rc;
   current->state = TASK_PAUSING;

   if (current->timeout == LLONG_MAX && rc->selectMode == KO) { 
      resourcePushWaitingTask(&(rc->waitNotFull), current); 
      RelRes(rc);
      pauseTask(thread);
      }

   RelRes(rc);

   current->sleep = getTick(OK, NULL) + 2000;
   addSleepingTask(thread, current);
   pauseTask(thread); 

LReturnPutRelease:
   RelRes(rc);

LReturnPut:
   current->waitingResource = NULL;
   return ret;
}
//*****************************************************************************
void resourceClose(register CThread *thread, register MyResource *res)
//********************************************************************
{
register MyTask *t;

   LockRes(res);

      res->oclosed = OTRUE;

      while((t = resourcePopTaskWaitingGet(res)) != NULL) awakeTask(thread, t);
      while((t = resourcePopTaskWaitingPut(res)) != NULL) awakeTask(thread, t);

   RelRes(res);
}
//*****************************************************************************
void N_Resource_close(register CThread *thread)
//******************************************
{
   resourceClose(thread, (MyResource *) TOP);
   DROP; 
}
//******************************************************************************


//******************************************************************************
// Console
//******************************************************************************

//******************************************************************************
int myConsoleSelect(MyResource *rc, vint mode)
//********************************************
{
#ifdef __OWIN__
   if (_kbhit() == 0) return 0;
   return 1;
#endif

#if defined(__OLNX32__) || defined(__OLNX64__)
int ch;

   ch = getchar();
   if (ch == EOF) return 0;
   ungetc(ch, stdin);
   return 1;
#endif


#if defined(__OMAC32__) || defined(__OMAC64__)
#define STDIN   0
fd_set fd;
struct timeval timeout;
int ret;

   timeout.tv_sec = 0;
   timeout.tv_usec = 0;
   FD_ZERO(&fd);
   FD_SET(STDIN, &fd);
   ret = select(STDIN+1, &fd, NULL, NULL, &timeout);
   return ret;
#endif
}

//*****************************************************************************
void N_Console_ekey(register CThread *thread)
//*******************************************
{
register MyConsole *console;

   console = (MyConsole *) TOP;

   if (myConsoleSelect((MyResource *) console, RECEIVE_MODE) == OK)
      { TOP = (Cell) OTRUE; } else { TOP = (Cell) OFALSE; }
}
//*****************************************************************************

#ifdef __OWIN__
//*****************************************************************************
Cell consoleGet(register CThread *thread, MyResource *rc, vint *perror)
//*********************************************************************
{
register MyConsole *console;
vint virt, ret = 0, virt2 = 0;
vint c = 0;
DWORD count = 0;
INPUT_RECORD rec = { 0 };

   console = (MyConsole *) rc;

   while(1) {
      if (myConsoleSelect((MyResource *) console, RECEIVE_MODE) == 0) break;
      ReadConsoleInput(STDIN, &rec, 1, &count);
      if (count == 0) break;
      if (rec.EventType != KEY_EVENT) continue;
      if (rec.Event.KeyEvent.bKeyDown == FALSE) continue;

      c    = rec.Event.KeyEvent.uChar.UnicodeChar;
      virt = rec.Event.KeyEvent.wVirtualKeyCode;

      // dbg("virt = %x, c = %c %d (virt2 = %x)\n", virt, c, c, virt2);
      if (virt == 0x10 || virt == 0x11 || virt == 0x12) { virt2 = virt; continue; }

      // If char is a CR on windows, translate to '\n' to be compatible
      if (c == '\r') c = '\n';
      if (virt2 == 0x11) c = virt;
      if (c != 0 ) { ret = (virt2 << 22) + c; } else { ret = virt << 22 ;}
      break;
      }

   if (ret == 0) { return NULL; }
   return (Cell) TOOINT(ret);
}
#endif

#ifdef __OLNX__
//*****************************************************************************
Cell consoleGet(register CThread *thread, MyResource *rc, vint *perror)
//*********************************************************************
{
register MyConsole *console;
vint virt = 0, ret = 0, size;
vint c = 0, code = 0, d, translate, i;

   console = (MyConsole *) rc;

   if (myConsoleSelect((MyResource *) console, RECEIVE_MODE) == 0) return NULL;

   c = getchar();
   // dbg("Code = %d\n", code);
   
   if (c == 127) { c = 8; }
   else if (c == 27) {
      c = 0;
      code = getchar();
      // dbg("Code = %d\n", code);
      if (code == EOF) { virt = 0; }
      else {
         virt = getchar();
         // dbg("virt = %d\n", virt);
         if (virt == 65)      { virt = 0x26; }
         else if (virt == 66) { virt = 0x28; }
         else if (virt == 67) { virt = 0x27; }
         else if (virt == 68) { virt = 0x25; }
         else if (virt >= 80 && virt <= 92) { virt = 0x70 + virt - 80; }
         else if (virt == 53) { virt = 0x21; getchar(); }
         else if (virt == 54) { virt = 0x22; getchar(); }
         else if (virt == 52) { virt = 0x23; getchar(); }
         }
      }
   else {
      if (c > 0x7F && c != (uchar) EOF) {
         if (c >= 0xC2 && c <= 0xDF)        { size = 1; translate = 0x0003080; }
         else { if (c >= 0xE0 && c <= 0xEF) { size = 2; translate = 0x00E2080; }
         else { if (c >= 0xF0 && c <= 0xF4) { size = 3; translate = 0x3C82008; }
         else { goto LConsoleError; } }}

         for(i = 1; i <= size; i++) {
            d = getchar();
            if (d < 0x80 || d > 0xBF) goto LConsoleError;
            if (i == 1) {
               if (size == 2) {
                  if (c == 0xE0 && d < 0xA0) goto LConsoleError;
                  if (c == 0xED && d > 0x9F) goto LConsoleError;
                  }
               if (size == 3) {
                  if (c == 0xF0 && d < 0x90) goto LConsoleError;
                  if (c == 0xF4 && d > 0x8F) goto LConsoleError;
                  }
               }
            c = (c << 6) + d;
            }
         c -= translate;
         }
      }

LConsoleEnd:
   virt &= 0x7F;
   ret = (virt << 22) + c;
   if (ret == 0) return NULL;
   return (Cell) TOOINT(ret);

LConsoleError:
   c = '?'; goto LConsoleEnd;
}
#endif

//*****************************************************************************
void N_Console_receiveTimeout(register CThread *thread)
//*****************************************************
{
MyTask *current;
vint error = 0;

   current = thread->currentTask;

   if (current->waitingResource == NULL) {
      CHECKSTACK(1)
      retrieveTimeout(thread, current, (MYINT) TOP);
      }

   TOP = resourceGet(thread, (MyResource *) mainConsole, consoleGet, &error);
}
//*****************************************************************************


//*****************************************************************************
// Channels
//*****************************************************************************

//*****************************************************************************
void N_Channel_initialize(register CThread *thread)
//************************************************
{
register MyChannel *ch;
register MYINT size, osize;

   // dbg("Creating channel : %x\n", ch);

   CHECKSTACK(1)
   ch = (MyChannel *) TOP;
   osize = (MYINT) SECOND; if (!ISINT(osize)) outOfMemory(thread);
   size = TOCINT(osize);
   if (size < 0 || size >= 0x10000000) outOfMemory(thread);

   ch->buf  = allocMem(thread, size * sizeof(Cell), NULL);
   ch->maxSize = size;
   ch->size  = 0;
   ch->first = 0;
   ch->last  = size - 1;

   ch->prev = NULL;
   ch->next = thread->channels; 
   if (ch->next != NULL) ch->next->prev = ch;
   thread->channels = ch;

   DROP2; 
}
//*****************************************************************************
void channelFinalize(register CThread *thread, MyObject *o)
//*********************************************************
{
register MyChannel *ch;

   ch = (MyChannel *) o;

   // dbg("Finalizing channel : %x\n", ch);
   if (ch->oclosed == OFALSE) resourceClose(thread, (MyResource *) ch);
   if (ch->buf != NULL) freeMem(thread, ch->buf);

   if (ch == thread->channels) {
         thread->channels = ch->next; 
         if (ch->next != NULL) ch->next->prev = NULL;
         }
      else { 
         ch->prev->next = ch->next; 
         if (ch->next != NULL) ch->next->prev = ch->prev;
         }
}
//*****************************************************************************
Cell channelGet(CThread *thread, MyResource *rc, vint *perror)
//************************************************************
{
register Cell e;
register MyChannel *ch;;

   ch = (MyChannel *) rc;
   if (ch->size != 0) {
      e = ch->buf[ch->first];
      ch->first = (ch->first + 1) % ch->maxSize;
      ch->size = ch->size - 1;
      return e;
      }

   return NULL;
}
//*****************************************************************************
void N_Channel_receiveTimeout(CThread *thread)
//*******************************************
{
MyTask *current;
vint error = 0;
Cell ret;

   current = thread->currentTask;

   if (current->waitingResource == NULL) {
      CHECKSTACK(2)
      retrieveTimeout(thread, current, (MYINT) SECOND);
      }

   ret = resourceGet(thread, (MyResource *) TOP, channelGet, &error);
   if (ret != NULL) { SECOND = ret;                  TOP = (Cell) OTRUE;  }
   else             { SECOND = (Cell) TOOINT(error); TOP = (Cell) OFALSE; }
}
//*****************************************************************************
int channelPut(CThread *thread, MyResource *rc, MyObject *o, vint *perror)
//************************************************************************
{
register MyChannel *ch;

   ch = (MyChannel *) rc;

   if (ch->size != ch->maxSize) {
      ch->last = (ch->last + 1) % ch->maxSize;
      ch->buf[ch->last] = (Cell) o;
      ch->size = ch->size + 1;
      return OK;
      }

   return KO;
}
//*****************************************************************************
void N_Channel_sendTimeout(register CThread *thread)
//*************************************************
{
register MyTask *current;
register MyObject *e;
vint error = 0;

   current = thread->currentTask;
   e = (MyObject *) THIRD;

   if (current->waitingResource == NULL) {
      CHECKSTACK(3)

      if (e == NULL) 
         sysExceptionFromC(thread, "Non null object required into #Channel>>send");

      checkImmutable(thread, e);
      retrieveTimeout(thread, current, (MYINT) SECOND);
      }

   if (resourcePut(thread, (MyResource *) TOP, e, channelPut, &error) == OK) 
        { THIRD = (Cell) OTRUE; DROP2; } 
   else { THIRD = (Cell) TOOINT(error); SECOND = (Cell) OFALSE; DROP }
}
//*****************************************************************************


#ifdef __TCP_ENABLE__

//******************************************************************************
// Sockets methods
//******************************************************************************

//******************************************************************************
int socketClose(register CThread *thread, register MyTCPSocket *s)
//****************************************************************
{
int ret = 0;

   // dbg("Closing sock %x, sock = %x %x\n", s, s->sock, s->sockv6);
   s->oclosed = OTRUE;
   if (s->sock == nil && s->sockv6 == nil) return 0; 

   if (s->sock != nil) { 
     ret = SockClose(s->sock);
     // dbg("ret1 = %d\n", ret);
     if (ret != 0) return ret; 
     s->sock = nil;
     }
   if (s->sockv6 != nil) { 
      ret = SockClose(s->sockv6); 
      // dbg("ret2 = %d\n", ret);
      if (ret != 0) return ret; 
      s->sockv6 = nil; 
      }
   resourceClose(thread, (MyResource *) s);
   return 0;
}
//*****************************************************************************
void sysSocketFinalize(register CThread *thread, MyObject *s)
//***********************************************************
{
   // dbg("Finalizing socket...%x, on class %s\n", s, objectClass(s)->name->buf);
   socketClose(thread, (MyTCPSocket *) s);
}
//******************************************************************************
void N_Socket_close(register CThread *thread)
//******************************************
{
   // dbg("Closing socket...\n");
   if (socketClose(thread, (MyTCPSocket *) TOP) != 0) sysExOS(thread);
   DROP;
}
//******************************************************************************
void N_Socket_initialize(register CThread *thread)
//***********************************************
{
register MyTCPSocket *s;

   s = (MyTCPSocket *) TOP;
   s->selectMode = OK;
   DROP; 
}
//******************************************************************************


//******************************************************************************
// Sockets Server methods
//******************************************************************************

//******************************************************************************
void N_SocketServer_initialize(register CThread *thread)
//******************************************************
{
register MyTCPSocketServer *ss;
int ret, yes = 1;
vint mode;
MYINT backlog;
MyString *port;
struct addrinfo *phints, *res, *p;
u_long blockMode = 0;

#ifdef __OWIN__
SOCKET desc = 0, descv4 = 0, descv6 = 0;
char *adrval = (char *) (&yes);
#endif

#ifdef __OLNX__
int desc = 0, descv4 = 0, descv6 = 0;
void *adrval = (void *) (&yes);
#endif

   CHECKSTACK(4)

   ss = (MyTCPSocketServer *) TOP;
   ss->selectMode = OK;

   port = (MyString *) FOURTH; 
   checkType(thread, port, String);
   ss->port = port;

   backlog = (MYINT) THIRD;
   checkTypeInteger(thread, backlog);
   ss->backlog = backlog;

   mode = (MYINT) SECOND;
   checkTypeInteger(thread, mode);

   ss->mode = mode;
   mode = TOCINT(mode);
   if (mode > MODE_IPALL) badParams(thread);

   // Load address
   phints = (struct addrinfo *) FORMAT;
   memset(phints, 0, sizeof(struct addrinfo));
   phints->ai_family = AF_UNSPEC;
   phints->ai_socktype = SOCK_STREAM;
   phints->ai_flags = AI_PASSIVE;

   if (getaddrinfo(NULL, (char *) port->buf, phints, &res) != 0) sysExOS(thread);

   for(p = res; p != NULL; p = p->ai_next) {
      if (p->ai_family == AF_INET  && ((mode & MODE_IPV4) == 0)) continue;
      if (p->ai_family == AF_INET6 && ((mode & MODE_IPV6) == 0)) continue;

      desc = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
      if (desc == INVALID_SOCKET) continue;

      if (setsockopt(desc, SOL_SOCKET, SO_REUSEADDR, adrval, sizeof(int)) == -1)
         sysExOS(thread);

      if (p->ai_family == AF_INET6) {
         if (setsockopt(desc, IPPROTO_IPV6, IPV6_V6ONLY, adrval, sizeof(int)) == -1)
            sysExOS(thread);
         }

      // Bind
      if (bind(desc, p->ai_addr, p->ai_addrlen) == -1) { close(desc); continue; }

      if (p->ai_family == AF_INET)  descv4 = desc;
      if (p->ai_family == AF_INET6) descv6 = desc;
      }

   freeaddrinfo(res);

   // Listen
   if ((mode & MODE_IPV4) && descv4 == 0)
       sysExceptionFromC(thread, "No IP V4 supported");
   else {
      ss->sock = descv4;
      if ( listen(descv4, TOCINT(backlog)) == SockError) sysExOS(thread);

      // Set new socket V4 as non blocking
      SocketSetNonBlocking(descv4, ret); 
      if (ret == SockError) sysExOS(thread);
      }

   if ((mode & MODE_IPV6) && descv6 == 0)
      sysExceptionFromC(thread, "No IPV6 supported");
   else {
      ss->sockv6 = descv6;
      if (listen(descv6, TOCINT(backlog)) == SockError) sysExOS(thread);

      // Set new socket V6 as non blocking 
      SocketSetNonBlocking(descv6, ret); 
      if (ret == SockError) sysExOS(thread);
      }

   // dbg("Socket ipv4 = %d\n", descv4);
   // dbg("Socket ipv6 = %d\n", descv6);
   ss->oclosed = OFALSE;
   DROP4;
}
//*******************************************************************************
Cell socketServerAccept(register CThread *thread, MyResource *rc, vint *perror)
//*****************************************************************************
{
struct sockaddr_storage *their_addr;
register MyTCPSocketServer *ss;
socklen_t sin_size;
MyTCPSocketClient *sc = NULL;
int sssock, sock= 0, ret = 0;
u_long blockMode = 0;

   if (rc->oclosed == OTRUE) { *perror = ERESCLOSED; return NULL; }

   their_addr = (struct sockaddr_storage *) FORMAT;
   ss = (MyTCPSocketServer *) rc;
   sin_size = sizeof(struct sockaddr_storage);
   memset(their_addr, 0, sin_size);

   sssock = ss->sock;
   if (sssock == 0) sssock = ss->sockv6;

   sock = accept(sssock, (struct sockaddr *) their_addr, &sin_size);

   if (sock == INVALID_SOCKET) {
      if (myerrno != CWOULDBLOCKA) { *perror = ESOCKACCEPT; }
      return NULL;
      }

   sc = (MyTCPSocketClient *) THIRD; 

   // Create new TCPSocketClient
   getnameinfo((struct sockaddr *) their_addr, sin_size, 
               (char *) sc->host->buf, 100, (char *) sc->remoteport->buf, 8, 
               NI_NUMERICHOST | NI_NUMERICSERV);

   sc->host->free1       += strlen((char *) sc->host->buf);
   sc->remoteport->free1 += strlen((char *) sc->remoteport->buf);

   // Set TCP_NODELAY on this socket.
   ret = 1;
   if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (void *) &ret, sizeof(ret)) != 0) 
      { *perror = ESOCKOPT; return NULL; }

   // Set new socket client as non blocking socket
   SocketSetNonBlocking(sock, ret); 
   if (ret == SockError) { *perror = ESOCKOPT; return NULL; }

   // dbg("New connection created...\n"); 
   sc->sock = sock;
   return (Cell) sc;
}
//******************************************************************************
void N_SocketServer_acceptTimeout(register CThread *thread)
//*********************************************************
{
register MyTask *current ;
vint error = 0;

   current = thread->currentTask;

   if (current->waitingResource == NULL) {
      CHECKSTACK(3)
      retrieveTimeout(thread, current, (MYINT) SECOND);
      }

   resourceGet(thread, (MyResource *) TOP, socketServerAccept, &error); 

   if (error == 0) { SECOND = (Cell) OTRUE; }
   else            { THIRD =  (Cell) TOOINT(error); SECOND = (Cell) OFALSE; }
   DROP;
}
//******************************************************************************


//******************************************************************************
// Sockets Client methods
//******************************************************************************

//******************************************************************************
Cell socketConnect(register CThread *thread, MyResource *rc, vint *perror)
//***********************************************************************
{
register MyTCPSocketClient *sc;
// Warning : this method can pause : set unused variables
int ret = 0, mode = 0;
int optval = 0;
struct addrinfo *p = NULL, *phints = NULL;
int desc = 0;

#ifdef __OWIN__
char *adrval = (char *) (&optval);
int optlen = sizeof(int);
#endif

#ifdef __OLNX__
void *adrval = (void *) (&optval);
socklen_t optlen = sizeof(int);
#endif


   sc = (MyTCPSocketClient *) rc;

   if (sc->results == NULL) {
      phints = (struct addrinfo *) FORMAT;
      memset(phints, 0, sizeof(struct addrinfo));
      mode = TOCINT(THIRD);
      if      (mode == MODE_IPV4) { phints->ai_family = AF_INET; }
      else if (mode == MODE_IPV6) { phints->ai_family = AF_INET6; }
      else                        { phints->ai_family = AF_UNSPEC; }

      phints->ai_socktype = SOCK_STREAM;

      // Retreive server addresses
      ret = getaddrinfo((char *) sc->host->buf, (char *) sc->remoteport->buf, phints, &(sc->results));
      if (ret != 0) { *perror = ESOCKHOST; return NULL; }
      sc->currentResult = sc->results;
      }
      
   // If connect was pending, check if connection is OK and return.
   if (sc->remainingSize != 0) {
      sc->remainingSize = 0;

      ret = getsockopt(sc->sock, SOL_SOCKET, SO_ERROR, adrval, &optlen);
      if (ret < 0) {
         if (sc->results != NULL) freeaddrinfo(sc->results);
         *perror = ESOCKCONNECT;
         return NULL;
         }

      // If connect is still pending
      // if (optval == EALREADY || optval == EINPROGRESS) return NULL;

      // If connect is OK
      if (optval == 0) { 
         if (sc->results != NULL) freeaddrinfo(sc->results);
         return (Cell) sc;
         }

      // Check next adress
      ret = optval;
      sc->currentResult = sc->currentResult->ai_next;
      }

   // Create client socket for current address
   while(sc->currentResult != NULL) {
      p = sc->currentResult;

      desc = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
      if (desc == INVALID_SOCKET) { sc->currentResult = p->ai_next; continue; }

      // Now, try to connect to current address
      ret = connect(desc, p->ai_addr, p->ai_addrlen);

      if (ret != SockError) {
         // TO DO : if (sc->port == OZERO) sc->port = TOOINT(ntohs(sin.sin_port));
         freeaddrinfo(sc->results);
         sc->sock = desc;
         sc->results = sc->currentResult = NULL;
         sc->oclosed = OFALSE;
         return (Cell) sc;
         }

      ret = myerrno;
      if (ret == CWOULDBLOCK) { sc->remainingSize = 1; return NULL; }

      sc->currentResult = p->ai_next;
      }

   // All addresses are KO, leaving with KO
   freeaddrinfo(sc->results);
   *perror = ESOCKCONNECT;
   return NULL;
}
//******************************************************************************
void N_SocketClient_connectMode(register CThread *thread)
//*******************************************************
{
register MyTask *current;
register MyTCPSocketClient *sc;
u_long blockMode = 0;
int retnb = 0;
vint error = 0;

   current = thread->currentTask;
   sc = (MyTCPSocketClient *) TOP;

   if (current->waitingResource == NULL) {
      CHECKSTACK(3)
      sc->results = sc->currentResult = NULL;
      sc->remainingSize = 0;
      retrieveTimeout(thread, current, (MYINT) SECOND);
      }
   
   resourceGet(thread, (MyResource *) sc, socketConnect, &error);

   if (error == 0) {
      SocketSetNonBlocking(sc->sock, retnb); 
      if (retnb == SockError) { error = ESOCKCONNECT; }
      else { THIRD = (Cell) OTRUE; DROP2; return; }
      }

   THIRD = (Cell) TOOINT(error); SECOND = (Cell) OFALSE; DROP;
}
//*****************************************************************************
Cell socketClientRead(register CThread *thread, MyResource *rc, vint *perror)
//***************************************************************************
{
register MyTCPSocketClient *sc;
register MyBuffer *buf;
int ret = 0;
vint size;

   // dbg("Receiving...\n");
   sc = (MyTCPSocketClient *) rc;
   buf = (MyBuffer *) FOURTH; 
   size = TOCINT(THIRD); 

   while(1) {
      ret = recv(sc->sock, (char *) buf->free1, size, 0);
      // dbg("receiving ret = %d\n", ret);

      if (ret < 0) {
         ret = myerrno;
         if (ret == COMPINTR) continue; 
         if (ret != CWOULDBLOCK) *perror = ESOCKREAD;
         return NULL;
         }

      buf->free1 += ret; *(buf->free1) = '\0';
      // dbg("Received = %s\n", buf->buf);
      return (Cell) TOOINT(ret);
      }
}
//******************************************************************************
void N_SocketClient_receiveTimeout(register CThread *thread)
//**********************************************************
{
register MyTask *current;
register MyTCPSocketClient *sc;
// Warning : this method can pause : set unused variables
MyBuffer *mb = NULL;
vint error = 0, osize = 0, len = 0;
Cell ret;

   current = thread->currentTask;
   sc = (MyTCPSocketClient *) TOP;

   if (current->waitingResource == NULL) {
      CHECKSTACK(4)

      if (sc->sock == nil) 
         sysExParamFromC(thread, "Can't #receive on not connected socket", sc);

      mb = (MyBuffer *) FOURTH; 
      checkType(thread, mb, Buffer);

      if (!(mb->objptr & MSKMUTABLE))
         { sysExceptionFromC(thread, "Mutable buffer required into #receive"); }

      osize = (vint) THIRD; 
      checkTypeInteger(thread, osize);

      len = TOCINT(osize);
      if (mb->alloc - mb->free1 <= len) bufAddChars(thread, mb, NULL, len);

      retrieveTimeout(thread, current, (MYINT) SECOND);
      }

   // dbg("C : Socket %x receive begin, timeout = %I64d, wait = %x\n",
      // sc, current->timeout, current->waitingResource);

   ret = resourceGet(thread, (MyResource *) sc, socketClientRead, &error);

   if (error == 0) { FOURTH = ret;                  THIRD = (Cell) OTRUE;  }
   else            { FOURTH = (Cell) TOOINT(error); THIRD = (Cell) OFALSE; }
   DROP2;
}
//*****************************************************************************
int socketClientSend(register CThread *thread,MyResource *rc,MyObject *o,vint *perror)
//****************************************************************************
{
register MyTCPSocketClient *sc;
int ret = 0;

   sc = (MyTCPSocketClient *) rc;
   // dbg("Sending...\n");

   while(sc->remainingSize != 0) {
      ret = send(sc->sock, (char *) sc->remainingBuffer, sc->remainingSize, 0);
      // dbg("Sending ret = %d / ", ret);

      if (ret < 0) {
         ret = myerrno;
         // dbg("sending errno = %d\n", ret);
         if (ret == COMPINTR) continue; 
         if (ret != CWOULDBLOCK) *perror = ESOCKREAD;
         return KO;
         } 
      sc->remainingBuffer += ret; sc->remainingSize -= ret;
      }
   // dbg("Sending OK\n");

   return OK;
}
//******************************************************************************
void N_SocketClient_sendTimeout(register CThread *thread)
//*******************************************************
{
register MyTCPSocketClient *sc;
register MyBuffer *b;
register MyTask *current;
int ret = 0;
vint error = 0;

   current = thread->currentTask;
   sc = (MyTCPSocketClient *) TOP;
   b = (MyBuffer *) THIRD; 

   if (current->waitingResource == NULL) {
      CHECKSTACK(3)

      if (sc->sock == nil) 
         sysExParamFromC(thread, "Can't #send to not connected socket", sc);

      retrieveTimeout(thread, current, (MYINT) SECOND);
      
      if (ISINTNIL(b) || (( b->objptr & MSKBUFFER ) == 0))
         sysExceptionFromC(thread, "Buffer required into #send");

      sc->remainingBuffer  = b->buf;
      sc->remainingSize    = b->free1 - b->buf;
      }

   ret = resourcePut(thread, (MyResource *)sc,(MyObject *) b,socketClientSend, &error);

   if (ret == OK) { THIRD = (Cell) OTRUE; DROP2; }
   else           { THIRD = (Cell) TOOINT(error); SECOND = (Cell) OFALSE; DROP; }
}
//******************************************************************************

#endif      // __TCP_ENABLE__

